import logging
from threading import Semaphore
from typing import Optional

from libqcanvas_clients.canvas import CanvasClient, CanvasClientConfig
from libqcanvas_clients.panopto import PanoptoClient, PanoptoClientConfig
from libqcanvas_clients.util.request_exceptions import ConfigInvalidError
from PySide6.QtCore import Qt, QUrl, Slot
from PySide6.QtGui import QDesktopServices, QIcon
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QErrorMessage,
    QLabel,
    QLineEdit,
    QMessageBox,
    QProgressBar,
    QVBoxLayout,
    QWidget,
)
from qasync import asyncSlot

import qcanvas.settings as settings
from qcanvas import icons
from qcanvas.util.layouts import GridItem, grid_layout_widget, layout
from qcanvas.util.url_checker import is_url

_logger = logging.getLogger(__name__)

_tutorial_url = (
    "https://www.iorad.com/player/2053777/Canvas---How-to-generate-an-access-token-"
)


class _InputRow:
    def __init__(
        self,
        *,
        label: str,
        initial_value: str,
        placeholder_text: Optional[str] = None,
        is_password: bool = False,
    ):
        self._label = QLabel(label)
        self._input = QLineEdit(initial_value)

        if placeholder_text is not None:
            self._input.setPlaceholderText(placeholder_text)

        if is_password:
            self._input.setEchoMode(QLineEdit.EchoMode.Password)

    def set_error(self, message: Optional[str]) -> None:
        self._input.setStyleSheet("QLineEdit { border: 1px solid red }")
        self._input.setToolTip(message)

    def clear_error(self) -> None:
        self._input.setStyleSheet(None)
        self._input.setToolTip(None)

    def grid_row(self) -> list[QWidget]:
        return [self._label, self._input]

    def disable(self) -> None:
        self._input.setEnabled(False)

    @property
    def enabled(self) -> bool:
        return self._input.isEnabled()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._input.setEnabled(value)

    @property
    def text(self) -> str:
        return self._input.text().strip()

    @property
    def url_text(self) -> str:
        url = self.text

        if not url.startswith("http"):
            return "https://" + url
        else:
            return url

    @property
    def is_valid_url(self) -> bool:
        return is_url(self.url_text)

    @property
    def is_empty(self) -> bool:
        return len(self.text) == 0


class SetupDialog(QDialog):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Configure QCanvas")
        self.setMinimumSize(550, 200)
        self.resize(550, 200)
        self.setWindowIcon(QIcon(icons.branding.main_icon))

        self._semaphore = Semaphore()

        self._canvas_url_box = _InputRow(
            label="Canvas URL",
            initial_value=settings.client.canvas_url,
            placeholder_text="https://instance.canvas.com",
        )
        self._panopto_url_box = _InputRow(
            label="Panopto URL",
            initial_value=settings.client.panopto_url,
            placeholder_text="https://instance.panopto.com",
        )
        self._canvas_api_key_box = _InputRow(
            label="Canvas API Key",
            initial_value=settings.client.canvas_api_key,
            is_password=True,
        )
        self._disable_panopto_checkbox = QCheckBox("Continue without Panopto")
        self._disable_panopto_checkbox.setChecked(settings.client.panopto_disabled)
        self._disable_panopto_checkbox.checkStateChanged.connect(
            self._disable_panopto_check_changed
        )
        self._panopto_url_box.enabled = not settings.client.panopto_disabled
        self._button_box = self._setup_button_box()
        self._waiting_indicator = self._setup_progress_bar()

        self.setLayout(
            layout(
                QVBoxLayout,
                grid_layout_widget(
                    [
                        self._canvas_url_box.grid_row(),
                        self._canvas_api_key_box.grid_row(),
                        self._panopto_url_box.grid_row(),
                        [
                            GridItem(
                                self._disable_panopto_checkbox,
                                col_span=2,
                                alignment=Qt.AlignmentFlag.AlignRight,
                            )
                        ],
                    ]
                ),
                self._waiting_indicator,
                self._button_box,
            )
        )

    def _setup_button_box(self) -> QDialogButtonBox:
        box = QDialogButtonBox()
        box.addButton(QDialogButtonBox.StandardButton.Ok)
        box.addButton("Get a Canvas API Key", QDialogButtonBox.ButtonRole.HelpRole)

        box.accepted.connect(self._verify_settings)
        box.helpRequested.connect(self._help_requested)
        return box

    def _setup_progress_bar(self) -> QProgressBar:
        progress = QProgressBar()
        progress.setMaximum(0)
        progress.setMinimum(0)
        self._set_retain_size_when_hidden(progress)
        progress.hide()
        return progress

    def _set_retain_size_when_hidden(self, widget: QWidget) -> None:
        size_policy = widget.sizePolicy()
        size_policy.setRetainSizeWhenHidden(True)
        widget.setSizePolicy(size_policy)

    @asyncSlot()
    async def _verify_settings(self) -> None:
        if self._semaphore.acquire(False):
            try:
                self._clear_errors()

                if not self._check_all_inputs():
                    return

                self._waiting_indicator.setVisible(True)

                canvas_config = CanvasClientConfig(
                    api_token=self._canvas_api_key_box.text,
                    canvas_url=self._canvas_url_box.url_text,
                )

                if not await self._check_canvas_config(canvas_config):
                    return

                if self._panopto_enabled:
                    if not await self._check_panopto_config(canvas_config):
                        self._show_panopto_help()
                        return
            except Exception as e:
                _logger.warning("Checking config failed", exc_info=e)

                error_box = QErrorMessage(self)
                error_box.showMessage(f"Checking config failed: {e}")
            finally:
                self._waiting_indicator.setVisible(False)
                self._semaphore.release()

            _logger.debug("Credentials are a-ok!")
            self._save_and_close()
        else:
            _logger.debug("Validation already in progress")

    def _clear_errors(self) -> None:
        self._canvas_url_box.clear_error()
        self._panopto_url_box.clear_error()
        self._canvas_api_key_box.clear_error()

    def _check_all_inputs(self) -> bool:
        all_valid = True

        if not self._canvas_url_box.is_valid_url:
            all_valid = False
            self._canvas_url_box.set_error("Canvas URL is invalid")

        if self._canvas_api_key_box.is_empty:
            all_valid = False
            self._canvas_api_key_box.set_error("Canvas API key is empty")

        if self._panopto_enabled and not self._panopto_url_box.is_valid_url:
            all_valid = False
            self._panopto_url_box.set_error("Panopto URL is invalid")

        return all_valid

    async def _check_canvas_config(self, canvas_config: CanvasClientConfig) -> bool:
        try:
            await CanvasClient.verify_config(canvas_config)
            return True
        except ConfigInvalidError:
            self._canvas_api_key_box.set_error("Canvas API key is invalid")
            return False

    async def _check_panopto_config(self, canvas_config: CanvasClientConfig) -> bool:
        client = CanvasClient(canvas_config)
        try:
            await PanoptoClient.verify_config(
                PanoptoClientConfig(panopto_url=self._panopto_url_box.url_text),
                client,
            )
            return True
        except ConfigInvalidError:
            return False
        finally:
            await client.aclose()

    def _show_panopto_help(self):
        msg = QMessageBox(
            QMessageBox.Icon.Information,
            "Panopto Authentication",
            "In order for QCanvas to use Panopto, you need to link your Panopto account to your Canvas account. "
            "A page will open in your web browser to do this when you click OK. It may ask you to sign into Canvas.\n\n"
            'Please tick "Remember my authorisation for this service" or QCanvas may not function correctly.\n\n'
            "QCanvas can't access anything entered in your browser.",
            QMessageBox.StandardButton.Ok,
            self,
        )
        msg.accepted.connect(
            self._open_panopto_login, Qt.ConnectionType.SingleShotConnection
        )
        msg.show()

    @Slot()
    def _open_panopto_login(self) -> None:
        url = QUrl(self._panopto_url_box.url_text)
        url.setPath("/Panopto/Pages/Auth/Login.aspx")
        url.setQuery("instance=Canvas&AllowBounce=true")
        QDesktopServices.openUrl(url)

    def _save_and_close(self) -> None:
        settings.client.canvas_url = self._canvas_url_box.url_text

        if self._panopto_enabled:
            settings.client.panopto_url = self._panopto_url_box.url_text
        else:
            settings.client.panopto_disabled = True

        settings.client.canvas_api_key = self._canvas_api_key_box.text

        self.accept()

    @Slot()
    def _help_requested(self) -> None:
        msg = QMessageBox(
            QMessageBox.Icon.Information,
            "Help",
            "An interactive tutorial will open in your browser when you click OK.\n\n"
            'Note that the "purpose" text doesn\'t matter and you can enter anything you want.\n\n'
            'You should also leave the "expires" item blank if you want the key to last forever.\n\n'
            "Don't share this key. You can revoke it at any time.",
            parent=self,
        )
        msg.accepted.connect(
            self._open_tutorial, Qt.ConnectionType.SingleShotConnection
        )
        msg.show()

    @Slot()
    def _open_tutorial(self) -> None:
        QDesktopServices.openUrl(QUrl(_tutorial_url))

    @Slot(Qt.CheckState)
    def _disable_panopto_check_changed(self, state: Qt.CheckState) -> None:
        self._panopto_url_box.enabled = state == Qt.CheckState.Unchecked

    @property
    def _panopto_enabled(self) -> bool:
        return self._disable_panopto_checkbox.checkState() == Qt.CheckState.Unchecked



import logging
from typing import Optional

from PySide6.QtCore import Slot
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu

import qcanvas.settings as settings

_logger = logging.getLogger(__name__)


class SyncOnStartOption(QAction):
    def __init__(self, parent: Optional[QMenu] = None):
        super().__init__("Sync on start", parent)
        self.setToolTip(
            "When enabled, synchronisation will start when the application is opened."
        )
        self.setCheckable(True)
        self.setChecked(settings.client.sync_on_start)
        self.triggered.connect(self._triggered)
        # self.set(icons.options.sync_on_start)

    @Slot()
    def _triggered(self) -> None:
        settings.client.sync_on_start = self.isChecked()



import logging
from typing import Sequence

from libqcanvas import db
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.sync.sync_receipt import SyncReceipt

from qcanvas.ui.course_viewer.tabs.file_tab.file_tree import FileTree
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class PagesFileTree(FileTree):
    def __init__(self, tree_name: str, *, resource_manager: ResourceManager):
        super().__init__(
            tree_name=f"{tree_name}.pages", resource_manager=resource_manager
        )

    def create_tree_items(
        self, data: db.Course, sync_receipt: SyncReceipt
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for group in data.modules:  # type: db.Module
            if not group.pages:
                continue

            # Init group_widget lazily to prevent creating pointless tree widgets
            group_widget: MemoryTreeWidgetItem | None = None
            items_in_group = set()

            for item in group.pages:
                resource_widgets = []

                for resource in item.resources:  # type: db.Resource
                    if resource.id not in items_in_group:
                        items_in_group.add(resource.id)

                        if group_widget is None:
                            group_widget = self._create_group_widget(
                                group, sync_receipt
                            )

                        resource_widgets.append(
                            self._create_resource_widget(resource, sync_receipt)
                        )

                if len(resource_widgets) > 0:
                    group_widget.addChildren(resource_widgets)

            if group_widget is not None:
                widgets.append(group_widget)

        return widgets



import logging
import re

from bs4 import BeautifulSoup, ResultSet

_logger = logging.getLogger(__name__)


def clean_up_html(html: str) -> str:
    html = re.sub(r"background-color:\s?initial;?", "", html)
    html = html.replace("&nbsp;", " ")

    doc = BeautifulSoup(html, "html.parser")

    # Remove all scripts and css
    _remove_tags(doc.find_all(["link", "script"]))
    # Remove font awesome icons (which don't load anyway)
    _remove_tags(doc.find_all(["span"], class_=["dp-icon-content"]))
    # Remove screen reader elements
    _remove_tags(doc.find_all(class_="screenreader-only"))

    return str(doc)


def _remove_tags(tags: ResultSet) -> None:
    for tag in tags:
        tag.decompose()



import logging
from typing import Optional

from libqcanvas_clients.canvas import CanvasClientConfig
from libqcanvas_clients.panopto import PanoptoClientConfig

from qcanvas.util import paths
from ._mapped_setting import BoolSetting, MappedSetting

_logger = logging.getLogger(__name__)


class _ClientSettings:
    settings = paths.client_settings()
    canvas_url: MappedSetting[Optional[str]] = MappedSetting(default=None)
    canvas_api_key: MappedSetting[Optional[str]] = MappedSetting(default=None)
    panopto_url: MappedSetting[Optional[str]] = MappedSetting(default=None)
    panopto_disabled = BoolSetting(default=False)
    quick_sync_enabled = BoolSetting(default=False)
    sync_on_start = BoolSetting(default=False)
    download_new_resources = BoolSetting(default=False)
    download_new_videos = BoolSetting(default=False)

    @property
    def canvas_config(self) -> CanvasClientConfig:
        return CanvasClientConfig(
            api_token=self.canvas_api_key, canvas_url=self.canvas_url
        )

    @property
    def panopto_config(self) -> Optional[PanoptoClientConfig]:
        """
        Generates a panopto client config. If panopto is disabled, it returns None.
        """

        if self.panopto_disabled:
            _logger.debug("Panopto is disabled")
            return None
        else:
            return PanoptoClientConfig(panopto_url=self.panopto_url)



import asyncio
import logging
import sys

from PySide6.QtCore import Qt
from PySide6.QtGui import QGuiApplication
from PySide6.QtWidgets import QApplication
from qasync import QEventLoop

import qcanvas.backend_connectors.qcanvas_task_master as task_master
import qcanvas.settings as settings
from libqcanvas.qcanvas import QCanvas
from qcanvas.backend_connectors import FrontendResourceManager
from qcanvas.theme import app_theme
from qcanvas.ui.qcanvas_window import QCanvasWindow
from qcanvas.ui.setup import SetupDialog, setup_checker
from qcanvas.util import paths, runtime

_logger = logging.getLogger(__name__)
app = QApplication(sys.argv)


async def setup_database() -> QCanvas[FrontendResourceManager]:
    _qcanvas = QCanvas[FrontendResourceManager](
        canvas_config=settings.client.canvas_config,
        panopto_config=settings.client.panopto_config,
        storage_path=paths.data_storage(),
        resource_manager_class=FrontendResourceManager,
    )

    await _qcanvas.database.upgrade()
    await _qcanvas.init()

    return _qcanvas


def run_setup():
    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set, Qt.ConnectionType.SingleShotConnection)

    async def coro():
        setup_window = SetupDialog()
        setup_window.rejected.connect(lambda: exit())
        setup_window.show()
        await app_close_event.wait()

    asyncio.run(coro(), loop_factory=QEventLoop)


def launch():
    if runtime.is_running_as_flatpak:
        QGuiApplication.setDesktopFileName("io.github.qcanvas.QCanvasApp")

    app.setApplicationName("QCanvas")

    task_master.register()
    app_theme.theme = settings.ui.theme

    if setup_checker.needs_setup():
        run_setup()

    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set, Qt.ConnectionType.SingleShotConnection)

    async def async_main():
        _qcanvas = await setup_database()

        _main_window = QCanvasWindow(_qcanvas)
        _main_window.show()
        await app_close_event.wait()

    asyncio.run(async_main(), loop_factory=QEventLoop)



import logging

from PySide6.QtCore import Qt
from libqcanvas import db
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from libqcanvas.util import as_local
from PySide6.QtWidgets import (
    QLabel,
    QLayout,
    QDockWidget,
    QMainWindow,
)
from typing_extensions import override

import qcanvas.util.ui_tools as ui
from qcanvas.backend_connectors import FrontendResourceManager
from .assignment_tree import AssignmentTree
from qcanvas.ui.qml_components import CommentsPane, AttachmentsPane
from qcanvas.ui.course_viewer.tabs.content_tab import ContentTab
from qcanvas.ui.course_viewer.tabs.constants import (
    date_strftime_format,
)

_logger = logging.getLogger(__name__)


class AssignmentTab(ContentTab):
    def __init__(
        self,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: FrontendResourceManager,
    ):
        # must be before super init, otherwise _setup_layout will be called before it is initialised
        self._main_container = QMainWindow()

        super().__init__(
            explorer=AssignmentTree.create_from_receipt(
                course, sync_receipt=sync_receipt
            ),
            title_placeholder_text="No assignment selected",
            downloader=downloader,
        )

        # fixme: can't figure out how to get the panes to have the right size without showing them when nothing is selected
        self._comments_pane = CommentsPane(downloader)
        self._comments_dock = ui.dock_widget(
            title="Comments",
            name="comments_dock",
            widget=self._comments_pane,
            min_size=ui.size(150, 150),
            features=QDockWidget.DockWidgetFeature.DockWidgetMovable,
            hide=False,
        )

        self._submission_files_pane = AttachmentsPane(downloader)
        self._submission_files_dock = ui.dock_widget(
            title="Submitted Files",
            name="sub_files_dock",
            widget=self._submission_files_pane,
            min_size=ui.size(150, 150),
            features=QDockWidget.DockWidgetFeature.DockWidgetMovable,
            hide=False,
        )

        self._main_container.setCentralWidget(self._viewer)
        self._main_container.addDockWidget(
            Qt.DockWidgetArea.RightDockWidgetArea, self._submission_files_dock
        )
        self._main_container.addDockWidget(
            Qt.DockWidgetArea.RightDockWidgetArea, self._comments_dock
        )

        self._main_container.resizeDocks(
            [self._submission_files_dock, self._comments_dock],
            [350, 350],
            Qt.Orientation.Horizontal,
        )
        self._main_container.resizeDocks(
            [self._submission_files_dock],
            [200],
            Qt.Orientation.Vertical,
        )

        self._due_date_label = QLabel("")
        self._score_label = QLabel("")
        self.enable_info_grid()

    @override
    def _setup_layout(self) -> None:
        super()._setup_layout()
        self.content_grid.replaceWidget(
            self._viewer,
            self._main_container,
        )

        self.content_grid.setColumnStretch(0, 1)
        self.content_grid.setColumnStretch(1, 3)

    @override
    def setup_info_grid(self) -> QLayout:
        return ui.form_layout(
            {"Due": self._due_date_label, "Score": self._score_label},
        )

    # fixme: kind of a misleading name? it's not just updating the info "grid" anymore
    @override
    def update_info_grid(self, assignment: db.Assignment) -> None:
        if assignment.due_date is not None:
            due_text = as_local(assignment.due_date).strftime(date_strftime_format)
        else:
            due_text = "No due date"

        self._due_date_label.setText(due_text)

        last_submission = assignment.submissions[-1] if assignment.submissions else None
        submission_score = "-"

        if last_submission and last_submission.score:
            submission_score = last_submission.score

        self._score_label.setText(
            f"<b>{submission_score}</b>/{assignment.max_score or '?'}"
        )

        if last_submission and last_submission.attachments:
            self._submission_files_pane.load_files(last_submission.attachments)
            self._submission_files_dock.show()
        else:
            self._submission_files_pane.clear_files()
            self._submission_files_dock.hide()

        if last_submission and last_submission.comments:
            self._comments_pane.load_comments(last_submission.comments)
            self._comments_dock.show()
        else:
            self._comments_pane.clear_comments()
            self._comments_dock.hide()

    @override
    def _show_blank(self) -> None:
        super()._show_blank()

        self._comments_dock.hide()
        self._submission_files_dock.hide()



from PySide6.QtCore import Signal, Slot, QUrl
from PySide6.QtWidgets import QWidget
from libqcanvas.util import remove_unwanted_whitespaces, as_local
from qasync import asyncSlot

from qcanvas.backend_connectors import FrontendResourceManager
from .qml_bridge_types import Attachment, Comment
from libqcanvas import db
import logging
from .qml_pane import QmlPane

_logger = logging.getLogger(__name__)


class CommentsPane(QmlPane):
    attachment_opened = Signal(str)

    def __init__(
        self, downloader: FrontendResourceManager, parent: QWidget | None = None
    ):
        super().__init__(qml_path=QUrl("qrc:/qml/CommentsPane.qml"), parent=parent)
        self._downloader = downloader
        self._attachments: dict[str, db.Resource] = {}
        self._qattachments: dict[str, Attachment] = {}

        # Add context objects before we load the view
        self.ctx["comments"] = []
        self.load_view()

        self._downloader.download_finished.connect(self._download_updated)
        self._downloader.download_failed.connect(self._download_updated)

    def clear_comments(self) -> None:
        self.ctx["comments"] = []
        self._attachments.clear()
        self._qattachments.clear()

    def load_comments(self, comments: list[db.SubmissionComment]) -> None:
        qcomments = []

        self.parent().setWindowTitle(f"Comments ({len(comments)})")

        for comment in comments:
            attachments = []

            for attachment in comment.attachments:
                qattachment = Attachment(
                    file_name=attachment.file_name,
                    resource_id=attachment.id,
                    download_state=attachment.download_state,
                )
                qattachment.opened.connect(self._on_attachment_opened)
                attachments.append(qattachment)

                self._attachments[attachment.id] = attachment
                self._qattachments[attachment.id] = qattachment

            qcomments.append(
                Comment(
                    body=remove_unwanted_whitespaces(comment.body),
                    author=comment.author,
                    date=as_local(comment.creation_date).strftime("%Y-%m-%d %H:%M"),
                    attachments=attachments,
                    parent=self,
                )
            )

        self.ctx["comments"] = qcomments

    @asyncSlot(str)
    async def _on_attachment_opened(self, resource_id: str) -> None:
        if resource_id in self._attachments:
            await self._downloader.download_and_open(self._attachments[resource_id])
        else:
            _logger.warning(
                "User opened an attachment that doesn't belong to any comment! id=%s",
                resource_id,
            )

    @Slot(db.Resource)
    def _download_updated(self, resource: db.Resource) -> None:
        if resource.id in self._qattachments:
            self._qattachments[resource.id].download_state = resource.download_state



from PySide6.QtQml import QQmlContext


class ContextDict:
    def __init__(self, context: QQmlContext):
        self._context = context

    def __getitem__(self, item):
        return self._context.contextProperty(item)

    def __setitem__(self, key, value):
        self._context.setContextProperty(key, value)



import logging

from PySide6.QtCore import QSettings

_logger = logging.getLogger(__name__)


class MappedSetting[T]:
    """
    Acts as a proxy for a named value in a QSettings object.
    Stores the value in memory when initialised and updates it accordingly, to protect it from changes on disk.
    """

    def __init__(self, default: T | None = None):
        self.settings_object: QSettings
        self.default = default
        self.value = None

    def __get__(self, instance, owner) -> T:
        return self.value

    def __set__(self, instance, value) -> None:
        self.value = value
        self._write(value)

    def __set_name__(self, owner, name) -> None:
        if hasattr(owner, "settings") and isinstance(owner.settings, QSettings):
            self.settings_object = owner.settings
        else:
            raise AttributeError(
                "Expected owner object to have a 'settings' attribute of type QSettings"
            )

        self.setting_name = name
        self.value = self._read()

    def _read(self) -> object:
        return self.settings_object.value(self.setting_name, self.default)

    def _write(self, value: object) -> None:
        self.settings_object.setValue(self.setting_name, value)


class BoolSetting(MappedSetting[bool]):
    def __init__(self, default: bool = False):
        super().__init__(default)

    # @override
    def _read(self) -> bool:
        try:
            # noinspection PyTypeChecker
            value: str = super()._read()

            if isinstance(value, bool):
                return value

            return value.lower() == "true"
        except Exception as e:
            _logger.error("Could not read setting", exc_info=e)
            return self.default

    # @override
    def _write(self, value: object) -> None:
        if not isinstance(value, bool):
            raise TypeError()

        super()._write(str(value).lower())



from PySide6.QtCore import QUrl
from PySide6.QtQuick import QQuickView
from PySide6.QtWidgets import QGroupBox, QWidget

from qcanvas.theme import app_theme
from qcanvas.util.context_dict import ContextDict
import qcanvas.util.ui_tools as ui


class QmlPane(QGroupBox):
    def __init__(self, qml_path: QUrl, parent: QWidget | None = None):
        super().__init__(parent)
        self.qview = QQuickView(parent)
        self._qml_path = qml_path
        self.ctx = ContextDict(self.qview.rootContext())
        self.ctx["appTheme"] = app_theme

        self.setLayout(ui.hbox(QWidget.createWindowContainer(self.qview, self)))

    def load_view(self):
        self.qview.setSource(self._qml_path)



import logging
from typing import Sequence

from libqcanvas import db
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtWidgets import QHeaderView

from qcanvas import icons
from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.tree_widget_data_item import TreeWidgetDataItem

_logger = logging.getLogger(__name__)


class MailTree(ContentTree[db.Course]):
    def __init__(self, course_id: str):
        super().__init__(
            tree_name=f"course.{course_id}.mail",
            emit_selection_signal_for_type=db.Message,
        )

        self.ui_setup(
            header_text=["Subject", "Sender"],
            max_width=500,
            min_width=300,
            indentation=20,
            alternating_row_colours=True,
        )

        self.set_columns_resize_mode(
            [QHeaderView.ResizeMode.Stretch, QHeaderView.ResizeMode.ResizeToContents]
        )

    def create_tree_items(
        self, course: db.Course, sync_receipt: SyncReceipt
    ) -> Sequence[TreeWidgetDataItem]:
        widgets = []

        for message in course.messages:  # type: db.Message
            message_widget = self._create_mail_widget(message, sync_receipt)
            message_widget.setIcon(0, icons.tree_items.mail)
            widgets.append(message_widget)

        return widgets

    def _create_mail_widget(
        self, message: db.Message, sync_receipt: SyncReceipt
    ) -> TreeWidgetDataItem:
        message_widget = TreeWidgetDataItem(
            id=message.id,
            data=message,
            strings=[message.name, message.sender_name],
        )

        if sync_receipt.was_updated(message):
            self.mark_as_unseen(message_widget)

        return message_widget



import logging
from typing import Iterable, NamedTuple

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QGridLayout, QLayout, QWidget

_logger = logging.getLogger(__name__)


class GridItem(NamedTuple):
    widget: QWidget
    col_span: int = 1
    row_span: int = 1
    alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignLeft


def layout_widget[T](layout_type: type[T], *items: QWidget, **kwargs) -> QWidget:
    widget = QWidget()
    widget.setLayout(layout(layout_type, *items, **kwargs))
    return widget


def layout[T](layout_type: type[T], *items: QWidget | QLayout, **kwargs) -> T:
    result_layout: QLayout = layout_type(**kwargs)

    for item in items:
        if isinstance(item, QLayout):
            result_layout.addItem(item)
        else:
            result_layout.addWidget(item)

    return result_layout


def grid_layout_widget(grid: Iterable[Iterable[QWidget | GridItem]]) -> QWidget:
    widget = QWidget()
    widget.setLayout(grid_layout(grid))
    return widget


def grid_layout(grid: Iterable[Iterable[QWidget | GridItem]]) -> QGridLayout:
    result_layout = QGridLayout()

    for row, row_list in enumerate(grid):
        for col, item in enumerate(row_list):
            if isinstance(item, GridItem):
                result_layout.addWidget(
                    item.widget, row, col, item.row_span, item.col_span, item.alignment
                )
            else:
                result_layout.addWidget(item, row, col)

    return result_layout



import logging

from libqcanvas import db
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtCore import Slot
from PySide6.QtWidgets import (
    QGridLayout,
    QLabel,
    QWidget,
    QLayout,
)

from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.tabs.resource_rich_browser import ResourceRichBrowser

import qcanvas.util.ui_tools as ui

_logger = logging.getLogger(__name__)


class ContentTab(QWidget):
    @classmethod
    def create_from_receipt[T: ContentTab](
        cls: T,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: ResourceManager,
    ) -> type[T]:
        return cls(course=course, sync_receipt=sync_receipt, downloader=downloader)

    def __init__(
        self,
        *,
        explorer: ContentTree[db.Course],
        title_placeholder_text: str,
        downloader: ResourceManager,
    ):
        super().__init__()
        self.content_grid = QGridLayout()
        self._placeholder_text = title_placeholder_text
        self._title_label = self._create_title_label()
        self._use_info_grid = False
        self._info_grid = QWidget()
        self._info_grid.hide()
        self._viewer = ResourceRichBrowser(downloader=downloader)
        self._explorer = explorer

        self.setLayout(self.content_grid)
        self._setup_layout()
        self._explorer.item_selected.connect(self._item_selected)

    def enable_info_grid(self) -> None:
        self._info_grid.setLayout(self.setup_info_grid())
        self._use_info_grid = True

    def _create_title_label(self) -> QLabel:
        return ui.label(
            self._placeholder_text,
            font=ui.font(point_size=12, bold=True),
            allow_truncation=True,
        )

    def setup_info_grid(self) -> QLayout:
        """
        Override this if you need an info grid
        """
        raise NotImplementedError()

    def _setup_layout(self) -> None:
        self.content_grid.addWidget(self._explorer, 0, 0, 4, 1)
        self.content_grid.addWidget(self._title_label, 0, 1)
        self.content_grid.addWidget(self._info_grid, 1, 1)
        self.content_grid.addWidget(self._viewer, 2, 1)

    def reload(self, course: db.Course, *, sync_receipt: SyncReceipt) -> None:
        self._explorer.reload(course, sync_receipt=sync_receipt)

    @Slot(object)
    def _item_selected(self, item: object) -> None:
        if isinstance(item, db.CourseContentItem):
            _logger.debug("Show %s", item.name)
            self._show_content(item)
        else:
            self._show_blank()

    def _show_content(self, item: db.CourseContentItem) -> None:
        self._title_label.setText(item.name)
        self._viewer.show_content(item)

        if self._use_info_grid:
            self._info_grid.show()
            self.update_info_grid(item)

    def update_info_grid(self, content: db.CourseContentItem) -> None:
        raise NotImplementedError()

    def _show_blank(self) -> None:
        self._title_label.setText(self._placeholder_text)
        self._viewer.show_blank(completely_blank=True)

        if self._use_info_grid:
            self._info_grid.hide()



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.8.1
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00\x03O\
(\
\xb5/\xfd`d\x0b-\x1a\x00\x16gm# q\xdb\
\xccq\xe1\xe6\xd2\xb7\xd0\xfa\x906\x94\xdc\x14\xf9\x1c\xdb\
D{\x809,S\x9e\x83P\x05\xa0\x01\xc0\x0d8\x0f\
b\x00d\x00a\x00\xfe\xfeq\xba\x11D\x8cit\xda\
\x9e\x89\x1e\xe5D\x92\xc9\xe1!\x22'\x9e\xcf\xb0\xd1\x84\
\x0cc\xae\xd3dK\xbf>e\x8b7#\xf6\xab\xdfC\
\xe8\x10\x14\xa1\x84I\x11s\x9e\xde\x1b\x7f\x22\x9d\xd3\x90\
#\xf9\x11\xe9:\x97\xbf\x97\x93\x9b\xf4\xcf\xe5u\xfc.\
_\x9f\xcf\x00\xd6*\xa9c\x02Hf=IIEj\
\xc2~(\x18\x01\x07\xc3\x04\xbe\x989N\x7f\xb1D\xaa\
u\xd1\x9fCY\xda\xfc\xdf\x9bT\xcf\xd0I \x7f\x85\
\x83T\x8e\xeb]\x8523\x99\x8dd\x18\xc8\x1c\x92\x22\
Z3\x0e\x01&\xc1t\x9e\x90\xb5~\xb2\xf6O\x82\xed\
\xf9\xa4\xa0c\x01eu\xac\x148\x9e\xbf\xf2[_q\
x\xa7\xaf\xa4\xe5\xebp(\x08\xfd?\xf9\xd1\x93\xb2\xf8\
\x810)j\xde\xbb\xe3\xf5\x95\xbc\xe2\x06\x0d\xc2A*\
\xd6\xe8\x89w\xed\x90\xf3\xc9'}\x8d0^o\xa9~\
\xfb3\x98\xe2\xf9\xf3Q\xb7!\x0a\xbav\x9e\xb6w\xfc\
\x14r-n\xa7J\x8d_\xf2\x18\xdc\x96T?a\x06\
\x9f?\xcb)B\x0a\xe4@&~u\x98\xf1\xa6\xe2\x87\
\xe3\xfbTP\xb3u\xa1<\xdd\xf4x\x9f7\x09\xa9`\
\xec\xd5\xc7\xb6[\xe2\x7ff\x0a\xa5z\xce\x02\xcc\x85!\
\xd3I*\xb6\xf4/\xe5\x06\xd8\x0e\x8f\x5c\xc8\xe7\x1d\xc7\
,\xc1g\xe9?\xb3\xa6a\xf3\xe3\x07#r\x8f\x07\x11\
=]\x17\x9da\x9f_\xfc\xe3\x82\xd7\x1f }\x18\x08\
\xa9\x9e\xdf\xbe`\xd7,Zo\x96_\x9f/\x8e\x18\xc0\
\xc7K\xa1\x89\x07\xc8%\xde\xb9F?\x86X\xa4rN\
\xb3\x07\xd3\xd5\xff\xd8\xb2\xfc\x1ay\xc7\xad\xf9\xba\x5c\x96\
y\x80\xab\xa0\x01\x1a33$2\x93\x14\xa5\xb0\x1c \
\x84\x18\xc4\x10V\x07\xa9\xd1^C\xdd\x1dfi\x00\x82\
\x17\xc5\x10\x8b\xaeWv\xbd\xa2\x90\xb7\xfc?\x96\x99V\
\x8b\xf4\xf2{\xc1.\x8f>\xf5\x1b\xd8\x1c\x87\xe8\x12\xab\
\xd1\xca\x19\xbd+!'\x7fz\x09\x08\xa95\xdbgM\
\xb1\xd4\xben\xb8\xfb\xc8tbv\xa6\x9f\x7f\xb4\xde\x01\
bP\xb2\x030\x89D\x13\x86\xfd\xe8r\xd9\x94-\xd4\
\x95\x06\xb7\x84;,<\xd5f\xfd\xe7\x01o-\xc7~\
[Q\x9a2\xf6\x17<\x83$_B\x83\xb8v\xd4\x22\
\xb4\xafz\xd6\x1c\x0c\x10\xff'4pro\xd5\xbe\xad\
\xac\xce\x08\x868\xfd l\x1aU\xabM\xa9\xe6h\xd6\
-\xc8!\x99\xfa\xca\x05h\xd7X?jY(\xb7\x0f\
o<\xae\xeb\xf8]\xf9ax@\xf3\xe7_qw\x8b\
\xe60\x16\xee\xfcy\xfb1\xdb\xf1\xa2\xe3\x96\xd2ud\
\x04\x95\xed\xd8\xb6\xd4}\xd9\xb9\xday:\xa1ETN\
\xd4m\x89\x9c\x16\x05c\x17x\xcb<\xf2\x1e\xbf\xb1[\
\x9c}\xd5e\xe5\xbd\xb6!\xf6\x0fmy7\xb7\xc2\xd2\
\x83\x18\x13#}g\xd7$\x05\xa0\xadxi\x96W-\
d'\xde\xa4\x00P\xc3\xfdt\xa5w@_\x1af\xb8\
i=\x11I\x83XF\x938Im\x99\x0a$\xa7n\
\xaa\xc2@\x01\x8b~\x88g%\xe5E\x16\xe4\x06\xf6F\
[j\x9e\x87\x5c\xcbt\x07\xb2\xfe\x83A \xcet\x93\
\xd1\xb3\xdc\x01\xbe.@a%S\xa6\xb9\x99}\x80G\
\xb0\xa5\xf9\x01M\x8d\x07\xbe\xa4\xfaY\xbe\x98\xb4\x10R\
\xc0\xe4\xe2\xfb)\x082\xaeJP\xeb\x0b\xb8>\
\x00\x00\x02\x04\
i\
mport QtQuick\x0a\x0aL\
istModel {\x0a    L\
istElement {\x0a   \
     file_name: \
\x22texas.pdf\x22\x0a    \
    resource_id:\
 \x221\x22\x0a        dow\
nload_state: \x22NO\
T_DOWNLOADED\x22\x0a  \
  }\x0a    ListElem\
ent {\x0a        fi\
le_name: \x22oh_no_\
what_a_terribly_\
long_file_name_i\
ts_not_like_some\
one_would_actual\
ly_do_this.pdf\x22\x0a\
        resource\
_id: \x222\x22\x0a       \
 download_state:\
 \x22FAILED\x22\x0a    }\x0a\
    ListElement \
{\x0a        file_n\
ame: \x22i was tran\
sported to anoth\
er world where j\
avascript doesn'\
t exist.cbz\x22\x0a   \
     resource_id\
: \x223\x22\x0a        do\
wnload_state: \x22D\
OWNLOADED\x22\x0a    }\
\x0a}\x0a\
\x00\x00\x00\xa3\
i\
mport QtQuick\x0a\x0aT\
hemedRectangle {\
\x0a    anchors.fil\
l: parent\x0a\x0a    C\
ommentsList {\x0a  \
      anchors.fi\
ll: parent\x0a     \
   model: commen\
ts\x0a        palet\
te: theme\x0a    }\x0a\
}\x0a\
\x00\x00\x03\xf9\
i\
mport QtQuick\x0a\x0aR\
ectangle {\x0a    p\
roperty Palette \
theme: getTheme(\
)\x0a    color: the\
me.base\x0a\x0a    fun\
ction getTheme()\
 {\x0a        if (a\
ppTheme.theme ==\
= \x22native\x22)\x0a    \
        return p\
alette\x0a        e\
lse if (appTheme\
.dark_mode === t\
rue)\x0a           \
 return darkThem\
e\x0a        else\x0a \
           retur\
n lightTheme\x0a   \
 }\x0a\x0a    Palette \
{\x0a        id: li\
ghtTheme\x0a       \
 base: \x22#f8f9fa\x22\
\x0a        window:\
 \x22#f8f9fa\x22\x0a     \
   midlight: \x22#f\
8f9fa\x22\x0a        a\
ccent: \x22#e02424\x22\
\x0a        link: a\
ccent\x0a        da\
rk: \x22#dadce0\x22\x0a  \
      text: \x22#4d\
5157\x22\x0a    }\x0a\x0a   \
 Palette {\x0a     \
   id: darkTheme\
\x0a        base: \x22\
#202124\x22\x0a       \
 window: \x22#f8f9f\
a\x22\x0a        midli\
ght: \x22#202124\x22\x0a \
       accent: \x22\
#e02424\x22\x0a       \
 link: accent\x0a  \
      dark: \x22#3f\
4042\x22\x0a        te\
xt: \x22#e4e7eb\x22\x0a  \
  }\x0a\x0a    Connect\
ions {\x0a        t\
arget: appTheme\x0a\
\x0a        functio\
n onThemeChanged\
() {\x0a           \
 updateTheme()\x0a \
       }\x0a\x0a      \
  function onDar\
kModeChanged() {\
\x0a            upd\
ateTheme()\x0a     \
   }\x0a\x0a        fu\
nction updateThe\
me() {\x0a         \
   theme = getTh\
eme()\x0a        }\x0a\
    }\x0a}\x0a\
\x00\x00\x00\xae\
i\
mport QtQuick\x0a\x0aT\
hemedRectangle {\
\x0a    anchors.fil\
l: parent\x0a\x0a    A\
ttachmentsList {\
\x0a        anchors\
.fill: parent\x0a  \
      model: sub\
mission_files\x0a  \
      palette: t\
heme\x0a    }\x0a}\x0a\
\x00\x00\x03\xeb\
i\
mport QtQuick\x0aim\
port QtQuick.Con\
trols\x0aimport QtQ\
uick.Layouts\x0a\x0aIt\
em {\x0a    propert\
y alias text: co\
ntent.text\x0a    p\
roperty int line\
Width: 3\x0a    pro\
perty color line\
Colour: palette.\
accent\x0a    prope\
rty alias conten\
t: content\x0a\x0a    \
Layout.fillWidth\
: true\x0a    heigh\
t: content.conte\
ntHeight\x0a    cli\
p: true\x0a\x0a    Ite\
m {\x0a        anch\
ors.fill: parent\
\x0a        height:\
 content.content\
Height\x0a\x0a        \
Rectangle {\x0a    \
        id: deco\
ration\x0a         \
   width: lineWi\
dth\x0a            \
color: lineColou\
r\x0a            ra\
dius: 1\x0a        \
    anchors {\x0a  \
              le\
ft: parent.left\x0a\
                \
top: parent.top\x0a\
                \
bottom: parent.b\
ottom\x0a          \
  }\x0a        }\x0a\x0a \
       Text/*Edi\
t*/ {\x0a          \
  id: content\x0a  \
          Layout\
.fillWidth: true\
\x0a            col\
or: palette.text\
\x0a            //r\
eadOnly: true\x0a  \
          //text\
Format: TextEdit\
.AutoText\x0a      \
      anchors {\x0a\
                \
left: decoration\
.right\x0a         \
       leftMargi\
n: 10\x0a          \
      right: par\
ent.right\x0a      \
      }\x0a        \
}\x0a    }\x0a}\x0a\
\x00\x00\x08C\
i\
mport QtQuick\x0aim\
port QtQuick.Lay\
outs\x0a\x0aItem {\x0a   \
 id: delegate\x0a  \
  height: childr\
enRect.height\x0a  \
  anchors {\x0a    \
    left: parent\
.left\x0a        ri\
ght: parent.righ\
t\x0a    }\x0a\x0a    Col\
umnLayout {\x0a    \
    spacing: 0\x0a\x0a\
        anchors \
{\x0a            le\
ft: parent.left\x0a\
            righ\
t: parent.right\x0a\
        }\x0a\x0a     \
   RowLayout {\x0a \
           heigh\
t: childrenRect.\
implicitHeight\x0a \
           Layou\
t.fillWidth: tru\
e\x0a            sp\
acing: 10\x0a\x0a     \
       Image {\x0a \
               f\
unction attachme\
ntIcon(downloadS\
tatus) {\x0a       \
             swi\
tch(downloadStat\
us)\x0a            \
        {\x0a      \
                \
  case \x22DOWNLOAD\
ED\x22:\x0a           \
                \
 return \x22qrc:///\
icons/universal/\
downloads/downlo\
aded.svg\x22;\x0a     \
                \
   case \x22FAILED\x22\
:\x0a              \
              re\
turn \x22qrc:///ico\
ns/universal/dow\
nloads/download_\
failed.svg\x22;\x0a   \
                \
     case \x22NOT_D\
OWNLOADED\x22:\x0a    \
                \
        return \x22\
qrc:///icons/uni\
versal/downloads\
/not_downloaded.\
svg\x22;\x0a          \
              de\
fault:\x0a         \
                \
   return \x22qrc:/\
//icons/universa\
l/downloads/unkn\
own.svg\x22\x0a       \
             }\x0a \
               }\
\x0a\x0a              \
  id: fileIcon\x0a \
               s\
ource: attachmen\
tIcon(modelData.\
download_state)\x0a\
                \
fillMode: Image.\
PreserveAspectFi\
t\x0a              \
  sourceSize.hei\
ght: 17\x0a        \
        sourceSi\
ze.width: 17\x0a   \
         }\x0a\x0a    \
        Text {\x0a \
               i\
d: file_name_tex\
t\x0a              \
  text: modelDat\
a.file_name\x0a    \
            wrap\
Mode: Text.WrapA\
tWordBoundaryOrA\
nywhere\x0a        \
        Layout.f\
illWidth: true\x0a \
               f\
ont.underline: t\
rue\x0a            \
    color: palet\
te.link\x0a\x0a       \
         MouseAr\
ea {\x0a           \
         id: tex\
tClickArea\x0a     \
               a\
nchors.fill: par\
ent\x0a            \
        cursorSh\
ape: Qt.Pointing\
HandCursor\x0a     \
               h\
overEnabled: tru\
e\x0a              \
  }\x0a            \
}\x0a        }\x0a\x0a   \
     Spacer {\x0a  \
          size: \
5\x0a            vi\
sible: index !==\
 count - 1\x0a     \
   }\x0a    }\x0a\x0a    \
Connections {\x0a  \
      target: te\
xtClickArea\x0a\x0a   \
     function on\
Clicked() {\x0a    \
        modelDat\
a.opened(modelDa\
ta.resource_id)\x0a\
        }\x0a    }\x0a\
}\x0a\
\x00\x00\x01\xdd\
\x0a\
\x0a/*\x0aThis is a UI\
 file (.ui.qml) \
that is intended\
 to be edited in\
 Qt Design Studi\
o only.\x0aIt is su\
pposed to be str\
ictly declarativ\
e and only uses \
a subset of QML.\
 If you edit\x0athi\
s file manually,\
 you might intro\
duce QML code th\
at is not suppor\
ted by Qt Design\
 Studio.\x0aCheck o\
ut https://doc.q\
t.io/qtcreator/c\
reator-quick-ui-\
forms.html for d\
etails on .ui.qm\
l files.\x0a*/\x0aimpo\
rt QtQuick\x0a\x0aList\
View {\x0a    id: v\
iew\x0a    model: C\
ommentsListModel\
 {}\x0a    delegate\
: CommentsListDe\
legate {}\x0a}\x0a\
\x00\x00\x08\xb6\
i\
mport QtQuick\x0a\x0aL\
istModel {\x0a    L\
istElement {\x0a   \
     body: \x22This\
 is a comment. G\
aze at it. Be am\
azed.\x22\x0a        a\
uthor: \x22I am Ste\
ve\x22\x0a        date\
: \x222001-01-01 36\
:00\x22\x0a        att\
achments: [\x0a    \
        ListElem\
ent {\x0a          \
      file_name:\
 \x22texas.pdf\x22\x0a   \
             res\
ource_id: \x221\x22\x0a  \
              do\
wnload_state: \x22N\
OT_DOWNLOADED\x22\x0a \
           },\x0a  \
          ListEl\
ement {\x0a        \
        file_nam\
e: \x22oh_no_what_a\
_terribly_long_f\
ile_name_its_not\
_like_someone_wo\
uld_actually_do_\
this.pdf\x22\x0a      \
          resour\
ce_id: \x222\x22\x0a     \
           downl\
oad_state: \x22FAIL\
ED\x22\x0a            \
},\x0a            L\
istElement {\x0a   \
             fil\
e_name: \x22i was t\
ransported to an\
other world wher\
e javascript doe\
sn't exist.cbz\x22\x0a\
                \
resource_id: \x223\x22\
\x0a               \
 download_state:\
 \x22DOWNLOADED\x22\x0a  \
          }\x0a    \
    ]\x0a    }\x0a    \
ListElement {\x0a  \
      body: \x22If \
Morbius has a mi\
llion fans, I am\
 one of them.\x5cnI\
f Morbius has 5 \
fans, I am one o\
f them.\x5cnIf Morb\
ius has one fan,\
 That one is me.\
\x5cnIf Morbius has\
 no fans, I am n\
o longer alive.\x5c\
nIf the world is\
 against Morbius\
, I am against t\
he world.\x5cnTill \
my last breath, \
I'll love Morbiu\
s (2022).\x22\x0a     \
   author: \x22Goku\
\x22\x0a        date: \
\x221942-16-34 06:6\
3\x22\x0a        attac\
hments: [\x0a      \
      ListElemen\
t {\x0a            \
    file_name: \x22\
morbius_x264_108\
0.mkv\x22\x0a         \
       resource_\
id: \x22blahblah\x22\x0a \
               d\
ownload_state: \x22\
NOT_DOWNLOADED\x22\x0a\
            }\x0a  \
      ]\x0a    }\x0a  \
  ListElement {\x0a\
        body: \x22a\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaaaaaaa\
aaaaaaaaaaa\x5cnbbb\
bbbbbbbbbbbbbbbb\
bbbbbbbbbbbbbbbb\
bbbbbbbbbbbbbbbb\
bbbbbbbbbbbbbbbb\
bbbbbbbbbbbbbbbb\
bbbbbbbbbbbbbbbb\
bbbbbbb\x22\x0a       \
 author: \x22Word W\
rap Test\x22\x0a      \
  date: \x222000-02\
-02\x22\x0a        att\
achments: []\x0a   \
 }\x0a    ListEleme\
nt {\x0a        bod\
y: \x22x264 forms t\
he core of many \
web video servic\
es, such as Yout\
ube, Facebook, V\
imeo, and Hulu. \
It is widely use\
d by television \
broadcasters and\
 ISPs.\x22\x0a        \
author: \x22BideoJA\
N Wordsblahblah\x22\
\x0a        date: \x22\
2027-07-28\x22\x0a    \
    attachments:\
 [\x0a            L\
istElement {\x0a   \
             fil\
e_name: \x22Ass1_SO\
MEDUDE_48.pdf\x22\x0a \
               r\
esource_id: \x22bla\
h_v345\x22\x0a        \
        download\
_state: \x22DOWNLOA\
DED\x22\x0a           \
 }\x0a        ]\x0a   \
 }\x0a}\x0a\
\x00\x00\x01\xe3\
\x0a\
\x0a/*\x0aThis is a UI\
 file (.ui.qml) \
that is intended\
 to be edited in\
 Qt Design Studi\
o only.\x0aIt is su\
pposed to be str\
ictly declarativ\
e and only uses \
a subset of QML.\
 If you edit\x0athi\
s file manually,\
 you might intro\
duce QML code th\
at is not suppor\
ted by Qt Design\
 Studio.\x0aCheck o\
ut https://doc.q\
t.io/qtcreator/c\
reator-quick-ui-\
forms.html for d\
etails on .ui.qm\
l files.\x0a*/\x0aimpo\
rt QtQuick\x0a\x0aList\
View {\x0a    id: v\
iew\x0a    model: A\
ttachmentsListMo\
del {}\x0a    deleg\
ate: Attachments\
ListDelegate {}\x0a\
}\x0a\
\x00\x00\x00N\
i\
mport QtQuick\x0a\x0aI\
tem {\x0a    proper\
ty real size\x0a   \
 height: size\x0a  \
  width: 1\x0a}\x0a\
"

qt_resource_name = b"\
\x00\x03\
\x00\x00x<\
\x00q\
\x00m\x00l\
\x00\x1b\
\x0d\x0f\x19\xfc\
\x00C\
\x00o\x00m\x00m\x00e\x00n\x00t\x00s\x00L\x00i\x00s\x00t\x00D\x00e\x00l\x00e\x00g\
\x00a\x00t\x00e\x00.\x00u\x00i\x00.\x00q\x00m\x00l\
\x00\x18\
\x04\x161<\
\x00A\
\x00t\x00t\x00a\x00c\x00h\x00m\x00e\x00n\x00t\x00s\x00L\x00i\x00s\x00t\x00M\x00o\
\x00d\x00e\x00l\x00.\x00q\x00m\x00l\
\x00\x10\
\x03\xe1\xc1\xfc\
\x00C\
\x00o\x00m\x00m\x00e\x00n\x00t\x00s\x00P\x00a\x00n\x00e\x00.\x00q\x00m\x00l\
\x00\x13\
\x08\xe5\xf4\x1c\
\x00T\
\x00h\x00e\x00m\x00e\x00d\x00R\x00e\x00c\x00t\x00a\x00n\x00g\x00l\x00e\x00.\x00q\
\x00m\x00l\
\x00\x13\
\x05\xa6\xc5|\
\x00A\
\x00t\x00t\x00a\x00c\x00h\x00m\x00e\x00n\x00t\x00s\x00P\x00a\x00n\x00e\x00.\x00q\
\x00m\x00l\
\x00\x14\
\x0c\x17\xd1<\
\x00D\
\x00e\x00c\x00o\x00r\x00a\x00t\x00e\x00d\x00T\x00e\x00x\x00t\x00.\x00u\x00i\x00.\
\x00q\x00m\x00l\
\x00\x1b\
\x0c\xf7\xd6\xbc\
\x00A\
\x00t\x00t\x00a\x00c\x00h\x00m\x00e\x00n\x00t\x00s\x00L\x00i\x00s\x00t\x00D\x00e\
\x00l\x00e\x00g\x00a\x00t\x00e\x00.\x00q\x00m\x00l\
\x00\x13\
\x0a.\xb3\xfc\
\x00C\
\x00o\x00m\x00m\x00e\x00n\x00t\x00s\x00L\x00i\x00s\x00t\x00.\x00u\x00i\x00.\x00q\
\x00m\x00l\
\x00\x15\
\x0cYK\xbc\
\x00C\
\x00o\x00m\x00m\x00e\x00n\x00t\x00s\x00L\x00i\x00s\x00t\x00M\x00o\x00d\x00e\x00l\
\x00.\x00q\x00m\x00l\
\x00\x16\
\x0dg\x00\x9c\
\x00A\
\x00t\x00t\x00a\x00c\x00h\x00m\x00e\x00n\x00t\x00s\x00L\x00i\x00s\x00t\x00.\x00u\
\x00i\x00.\x00q\x00m\x00l\
\x00\x0d\
\x01wq\xfc\
\x00S\
\x00p\x00a\x00c\x00e\x00r\x00.\x00u\x00i\x00.\x00q\x00m\x00l\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x0b\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x01\xf4\x00\x00\x00\x00\x00\x01\x00\x00#i\
\x00\x00\x01\x94S\xec~\x99\
\x00\x00\x00~\x00\x00\x00\x00\x00\x01\x00\x00\x05[\
\x00\x00\x01\x94T\xb8\x224\
\x00\x00\x00H\x00\x00\x00\x00\x00\x01\x00\x00\x03S\
\x00\x00\x01\x94S\xec~\x97\
\x00\x00\x00\xd0\x00\x00\x00\x00\x00\x01\x00\x00\x09\xff\
\x00\x00\x01\x94T\xb8Af\
\x00\x00\x00\xa4\x00\x00\x00\x00\x00\x01\x00\x00\x06\x02\
\x00\x00\x01\x94T\xc4\xc8\xa7\
\x00\x00\x01f\x00\x00\x00\x00\x00\x01\x00\x00\x16\xe7\
\x00\x00\x01\x94S\xec~\x98\
\x00\x00\x00\xfc\x00\x00\x00\x00\x00\x01\x00\x00\x0a\xb1\
\x00\x00\x01\x94S\xec~\x98\
\x00\x00\x01\x92\x00\x00\x00\x00\x00\x01\x00\x00\x18\xc8\
\x00\x00\x01\x94S\xec~\x98\
\x00\x00\x01*\x00\x00\x00\x00\x00\x01\x00\x00\x0e\xa0\
\x00\x00\x01\x94S\xec~\x97\
\x00\x00\x00\x0c\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x94S\xec~\x98\
\x00\x00\x01\xc2\x00\x00\x00\x00\x00\x01\x00\x00!\x82\
\x00\x00\x01\x94S\xec~\x97\
"


def qInitResources():
    QtCore.qRegisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


def qCleanupResources():
    QtCore.qUnregisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


qInitResources()



import logging
from typing import Literal

import qdarktheme
from PySide6.QtCore import QObject, Signal, Property, Slot
from PySide6.QtGui import QGuiApplication, Qt, QIcon
from PySide6.QtWidgets import QStyleFactory, QApplication

type Theme = Literal["native", "auto", "dark", "light"]

_logger = logging.getLogger(__name__)


class _AppTheme(QObject):
    themeChanged = Signal()
    darkModeChanged = Signal()

    def __init__(self):
        super().__init__()
        self._last_system_theme = QGuiApplication.styleHints().colorScheme()
        self._theme: Theme | None = None
        self._dark_mode: bool | None = None

        self.darkModeChanged.connect(self._set_icon_paths)
        QGuiApplication.styleHints().colorSchemeChanged.connect(
            self._on_system_theme_changed
        )

    @Property(bool, notify=darkModeChanged)
    def dark_mode(self) -> bool:
        assert self._theme is not None, "Theme has not been set"
        return self._dark_mode

    @Property(str, notify=themeChanged)
    def theme(self) -> Theme:
        assert self._theme is not None, "Theme has not been set"
        return self._theme

    @theme.setter
    def theme(self, value: str):
        value = ensure_theme_is_valid(value)

        if value != self._theme:
            self._update_theme(value)

    def _update_theme(self, theme: str):
        if theme is None or (theme == self._theme and theme not in ["native", "auto"]):
            return

        was_dark_mode = self._dark_mode

        if theme != "native":
            if theme == "auto":
                self._dark_mode = _is_system_using_dark_mode()
                selected_colour_scheme = "dark" if self._dark_mode else "light"
            else:
                self._dark_mode = theme == "dark"
                selected_colour_scheme = theme

            if was_dark_mode != self._dark_mode:
                qdarktheme.setup_theme(
                    selected_colour_scheme,
                    custom_colors={"primary": "e02424"},
                )

                QApplication.setStyle(QStyleFactory.create("Fusion"))
        else:
            self._dark_mode = _is_system_using_dark_mode()

        if theme != self._theme:
            self._theme = theme
            self.themeChanged.emit()

        if was_dark_mode != self._dark_mode:
            self.darkModeChanged.emit()

    @Slot()
    def _set_icon_paths(self):
        QIcon.setFallbackSearchPaths(
            [":icons/dark" if self._dark_mode else ":icons/light", ":icons/universal"]
        )

    @Slot()
    def _on_system_theme_changed(self, scheme: Qt.ColorScheme):
        if scheme != self._last_system_theme:
            self._last_system_theme = scheme
            self._update_theme(self._theme)


def _is_system_using_dark_mode() -> bool:
    return QGuiApplication.styleHints().colorScheme() == Qt.ColorScheme.Dark


def ensure_theme_is_valid(theme_name: str) -> Theme:
    if theme_name not in ["auto", "light", "dark", "native"]:
        return "auto"
    else:
        return theme_name


app_theme = _AppTheme()



# Unused
import logging

from PySide6.QtGui import QIcon

_logger = logging.getLogger(__name__)


class ThemeIcon:
    def __init__(self, theme_path: str):
        self._theme_path = theme_path
        self._icon = None

    @property
    def icon(self) -> QIcon:
        if self._icon is None:
            self._icon = QIcon.fromTheme(self._theme_path)

        return self._icon

    @property
    def theme_path(self) -> str:
        return self._theme_path

    def __hash__(self) -> int:
        return hash(self._theme_path)


class UniversalIcon(ThemeIcon):
    def __init__(self, theme_path: str, full_path: str):
        super().__init__(theme_path)
        self._full_path = full_path

    @property
    def full_path(self) -> str:
        return self._full_path

    def __hash__(self) -> int:
        return hash(self._theme_path) ^ hash(self.full_path)


AnyIcon = UniversalIcon | ThemeIcon



import logging
from abc import ABC

from libqcanvas import db
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtCore import QPoint, Qt, Slot
from PySide6.QtWidgets import QHeaderView, QMenu, QTreeWidgetItem

from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.tree_widget_data_item import (
    AnyTreeDataItem,
    TreeWidgetDataItem,
)
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem
from qcanvas.util.file_icons import icon_for_filename
from qcanvas.util.ui_tools import create_qaction

_logger = logging.getLogger(__name__)


class FileTree(ContentTree[db.Course], ABC):
    @classmethod
    def create_from_receipt(
        cls,
        course: db.Course,
        *,
        sync_receipt: SyncReceipt,
        resource_manager: ResourceManager,
    ) -> "FileTree":
        tree = cls(tree_name=course.id, resource_manager=resource_manager)
        tree.reload(course, sync_receipt=sync_receipt)
        return tree

    def __init__(self, tree_name: str, *, resource_manager: ResourceManager):
        super().__init__(tree_name, emit_selection_signal_for_type=object)
        self._resource_manager = resource_manager

        self.ui_setup(header_text=["File", "Date"])
        self.set_columns_resize_mode(
            [QHeaderView.ResizeMode.Stretch, QHeaderView.ResizeMode.ResizeToContents]
        )

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)

    def _create_group_widget(
        self, group: db.ContentGroup, sync_receipt: SyncReceipt
    ) -> MemoryTreeWidgetItem:
        group_widget = MemoryTreeWidgetItem(
            id=group.id, data=group, strings=[group.name]
        )

        group_widget.setFlags(Qt.ItemFlag.ItemIsEnabled)

        if sync_receipt.was_updated(group):
            self.mark_as_unseen(group_widget)

        return group_widget

    def _create_resource_widget(
        self, resource: db.Resource, sync_receipt: SyncReceipt
    ) -> QTreeWidgetItem:
        # fixme the reesource widget items shouls NOT be a memory widget item because they can't be collapsed, but
        #  mostly because the same file can appear in the tree multiple times in different places, which memory tree
        #  can NOT deal with!
        item_widget = TreeWidgetDataItem(
            id=resource.id,
            data=resource,
            strings=[resource.file_name, str(resource.discovery_date.date())],
        )
        item_widget.setIcon(
            0,
            icon_for_filename(resource.file_name),
        )
        item_widget.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)

        if sync_receipt.was_updated(resource):
            self.mark_as_unseen(item_widget)

        return item_widget

    @Slot(QPoint)
    def _context_menu(self, point: QPoint) -> None:
        item = self.itemAt(point)

        if isinstance(item, AnyTreeDataItem):
            menu = QMenu()
            create_qaction(
                name="Test",
                parent=menu,
                triggered=lambda: print(f"Clicked {item.extra_data.file_name}"),
            )
            menu.addAction("Another thing")

            menu.exec(self.mapToGlobal(point))



from typing import List, Optional

from PySide6.QtWidgets import QTreeWidgetItem

from qcanvas.ui.memory_tree import MemoryTreeWidgetItem


class TreeWidgetDataItem(QTreeWidgetItem):
    def __init__(
        self, id: str, data: Optional[object], strings: Optional[List[str]] = None
    ):
        super().__init__(strings)
        # Still needs ID because it is used to reselect the item
        self._id = id
        self.extra_data = data

    @property
    def id(self) -> str:
        return self._id


AnyTreeDataItem = TreeWidgetDataItem | MemoryTreeWidgetItem



import logging
from math import floor
from typing import Optional, Sequence

from libqcanvas import db
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.sync.sync_receipt import SyncReceipt, empty_receipt
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QLabel, QSizePolicy, QStackedWidget

from qcanvas import icons
from qcanvas.ui.course_viewer.course_viewer import CourseViewer
from qcanvas.theme import app_theme

_logger = logging.getLogger(__name__)


class _PlaceholderLogo(QLabel):
    """
    Automatically resizing logo icon for when no course is selected
    """

    def __init__(self):
        super().__init__()
        self._icon = icons.branding.logo_transparent
        self._old_width = -1
        self._old_height = -1
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Ignored)
        # Because we are using a pixmap for the icon, it will not get updated like a normal QIcon when the theme changes,
        # So we need to update it ourselves
        app_theme.darkModeChanged.connect(self._dark_mode_changed)

    def resizeEvent(self, event) -> None:
        self._update_image()

    @Slot()
    def _dark_mode_changed(self) -> None:
        self._update_image(force=True)

    def _update_image(self, force: bool = False) -> None:
        # Calculate the size of the logo as half of the width/height, but prevent it from becoming too large
        width = min(floor(self.width() * 0.5), 500)
        height = min(floor(self.height() * 0.5), 500)

        if force or (width != self._old_width and height != self._old_height):
            self._old_width = width
            self._old_height = height
            self.setPixmap(self._icon.pixmap(width, height))


class CourseViewerContainer(QStackedWidget):
    def __init__(self, downloader: ResourceManager):
        super().__init__()
        self._course_viewers: dict[str, CourseViewer] = {}
        self._downloader = downloader
        self._last_course_id: Optional[str] = None
        self._selected_course: Optional[db.Course] = None
        self._last_sync_receipt: SyncReceipt = empty_receipt()
        self._placeholder = _PlaceholderLogo()
        self.addWidget(self._placeholder)

    def show_blank(self) -> None:
        self._last_course_id = None
        self._selected_course = None
        self.setCurrentWidget(self._placeholder)

    def load_course(self, course: db.Course) -> None:
        if course.id not in self._course_viewers:
            viewer = CourseViewer(
                course=course,
                downloader=self._downloader,
                sync_receipt=self._last_sync_receipt,
            )
            self._course_viewers[course.id] = viewer
            self.addWidget(viewer)
        else:
            viewer = self._course_viewers[course.id]

        self.setCurrentWidget(viewer)
        self._selected_course = course
        self._last_course_id = course.id

    async def reload_all(
        self, courses: Sequence[db.Course], *, sync_receipt: SyncReceipt
    ) -> None:
        self._last_sync_receipt = sync_receipt
        for course in courses:
            if course.id in self._course_viewers:
                viewer = self._course_viewers[course.id]
                viewer.reload(course, sync_receipt=sync_receipt)

    @property
    def selected_course(self) -> Optional[db.Course]:
        return self._selected_course



import logging
from threading import BoundedSemaphore
from typing import Optional, Sequence

import httpx
from libqcanvas import db
from libqcanvas.database.data_monolith import DataMonolith
from libqcanvas.net.sync.sync_receipt import SyncReceipt, empty_receipt
from libqcanvas.qcanvas import QCanvas
from PySide6.QtCore import QUrl, Signal, Slot, Qt
from PySide6.QtGui import QDesktopServices, QKeySequence
from PySide6.QtWidgets import (
    QErrorMessage,
    QHBoxLayout,
    QMainWindow,
    QProgressDialog,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from qasync import asyncSlot

from qcanvas import icons
from qcanvas.backend_connectors import FrontendResourceManager
from qcanvas.settings import course_configs
from qcanvas.ui.course_viewer import CourseTree
from .course_viewer_container import CourseViewerContainer
from .options.auto_download_resources_option import (
    AutoDownloadResourcesMenu,
)
from .options.quick_sync_option import QuickSyncOption
from .options.sync_on_start_option import SyncOnStartOption
from .options.theme_selection_menu import ThemeSelectionMenu
from .status_bar_progress_display import StatusBarProgressDisplay
from qcanvas.util import auto_downloader
import qcanvas.settings as settings
from qcanvas.util.qurl_util import file_url
from qcanvas.util.ui_tools import create_qaction

_logger = logging.getLogger(__name__)


class QCanvasWindow(QMainWindow):
    _loaded = Signal()

    def __init__(self, _qcanvas: QCanvas[FrontendResourceManager]):
        super().__init__()

        self.setWindowTitle("QCanvas")
        self.setWindowIcon(icons.branding.main_icon)

        self._operation_semaphore = BoundedSemaphore()
        self._data: Optional[DataMonolith] = None
        self._qcanvas = _qcanvas

        self._course_tree = CourseTree()
        self._course_tree.item_selected.connect(self._on_course_selected)
        self._course_tree.course_renamed.connect(self._on_course_renamed)
        self._sync_button = QPushButton("Synchronise")
        self._sync_button.clicked.connect(self._synchronise_requested)
        self._course_viewer_container = CourseViewerContainer(
            self._qcanvas.resource_manager
        )

        self.setCentralWidget(self._setup_main_layout())
        self.setStatusBar(StatusBarProgressDisplay())
        self._setup_menu_bar()
        self._restore_window_position()

        self._loaded.connect(
            self._on_app_loaded, Qt.ConnectionType.SingleShotConnection
        )
        self._loaded.emit()

    def _setup_menu_bar(self) -> None:
        menu_bar = self.menuBar()
        app_menu = menu_bar.addMenu("Actions")

        create_qaction(
            name="Synchronise",
            shortcut=QKeySequence("Ctrl+S"),
            triggered=self._synchronise_requested,
            parent=app_menu,
            icon=icons.actions.sync,
        )

        create_qaction(
            name="Open downloads folder",
            shortcut=QKeySequence("Ctrl+D"),
            triggered=self._open_downloads_folder,
            parent=app_menu,
            icon=icons.actions.open_downloads,
        )

        create_qaction(
            name="Open Canvas in browser",
            shortcut=QKeySequence("Ctrl+O"),
            triggered=self._open_quick_auth_in_browser,
            parent=app_menu,
            icon=icons.actions.quick_login,
        )

        create_qaction(
            name="Mark all as seen",
            triggered=self._clear_new_items,
            parent=app_menu,
            icon=icons.actions.mark_all_read,
        )

        create_qaction(
            name="Quit",
            shortcut=QKeySequence("Ctrl+Q"),
            triggered=lambda: self.close(),
            parent=app_menu,
            icon=icons.actions.exit,
        )

        options_menu = menu_bar.addMenu("Options")
        options_menu.setToolTipsVisible(True)
        options_menu.addAction(QuickSyncOption(options_menu))
        options_menu.addAction(SyncOnStartOption(options_menu))
        options_menu.addMenu(AutoDownloadResourcesMenu(options_menu))
        options_menu.addMenu(ThemeSelectionMenu(options_menu))

    def _restore_window_position(self):
        if settings.ui.last_geometry is not None:
            self.restoreGeometry(settings.ui.last_geometry)
        else:
            self.resize(1000, 600)

        if settings.ui.last_window_state is not None:
            self.restoreState(settings.ui.last_window_state)

    def closeEvent(self, event):
        settings.ui.last_geometry = self.saveGeometry()
        settings.ui.last_window_state = self.saveState()

    def _setup_main_layout(self) -> QWidget:
        h_box = QHBoxLayout()

        h_box.addLayout(self._setup_course_column(), 1)
        h_box.addWidget(self._course_viewer_container, 5)

        widget = QWidget()
        widget.setLayout(h_box)
        return widget

    def _setup_course_column(self) -> QVBoxLayout:
        course_list_column = QVBoxLayout()
        course_list_column.addWidget(self._course_tree)
        course_list_column.addWidget(self._sync_button)

        return course_list_column

    @asyncSlot()
    async def _on_app_loaded(self) -> None:
        self._course_tree.reload(await self._get_terms(), sync_receipt=empty_receipt())

        if settings.client.sync_on_start:
            await self._synchronise()

    @asyncSlot()
    async def _synchronise_requested(self) -> None:
        await self._synchronise()

    async def _synchronise(self) -> None:
        if not self._operation_semaphore.acquire(False):
            _logger.debug("Sync operation already in progress")
            return

        try:
            self._sync_button.setText("Sync in progress...")
            receipt = await self._qcanvas.synchronise_canvas(
                quick_sync=settings.client.quick_sync_enabled
            )
            await self._reload(receipt)
        except Exception as e:
            _logger.warning("Sync failed", exc_info=e)
            error = QErrorMessage(self)

            if isinstance(e, ExceptionGroup):
                msg = "\n".join(str(ex) for ex in e.exceptions)
            elif isinstance(e, httpx.ConnectError):
                msg = "You may not be connected to the Internet"
            else:
                msg = str(e)

            error.setWindowTitle("An error has occurred in QCanvas")
            error.showMessage(
                msg + " - Please check the log for more details", "sync error"
            )
        finally:
            self._operation_semaphore.release()
            self._sync_button.setText("Synchronise")

        try:
            if settings.client.download_new_resources:
                # noinspection PyUnboundLocalVariable
                await auto_downloader.download_new_resources(
                    all_resources=await self._get_resources(),
                    receipt=receipt,
                    downloader=self._qcanvas.resource_manager,
                    parent_window=self,
                )
        except NameError:
            pass

    async def _reload(self, receipt: SyncReceipt) -> None:
        self._course_tree.reload(await self._get_terms(), sync_receipt=receipt)
        await self._course_viewer_container.reload_all(
            await self._get_courses(), sync_receipt=receipt
        )

    async def _get_resources(self) -> dict[str, db.Resource]:
        return (await self._qcanvas.load()).resources

    async def _get_terms(self) -> Sequence[db.Term]:
        return (await self._qcanvas.load()).terms

    async def _get_courses(self) -> Sequence[db.Course]:
        return (await self._qcanvas.load()).courses

    @Slot(db.Course)
    def _on_course_selected(self, course: Optional[db.Course]) -> None:
        if course is not None:
            self._course_viewer_container.load_course(course)
        else:
            self._course_viewer_container.show_blank()

    @asyncSlot(db.Course, str)
    async def _on_course_renamed(self, course: db.Course, new_name: str) -> None:
        _logger.debug("Rename %s -> %s", course.name, new_name)
        config = course_configs[course.id]
        config.nickname = new_name
        await config.save()

    @asyncSlot()
    async def _open_quick_auth_in_browser(self) -> None:
        opening_progress_dialog = QProgressDialog("Opening canvas", None, 0, 0, self)
        opening_progress_dialog.setWindowTitle("Please wait")
        opening_progress_dialog.show()

        try:
            open_url = QUrl(
                await self._qcanvas.canvas_client.get_temporary_session_url()
            )
            _logger.info(f"Opening URL {open_url}")
            QDesktopServices.openUrl(open_url)
        finally:
            opening_progress_dialog.close()

    @Slot()
    def _open_downloads_folder(self) -> None:
        directory = self._qcanvas.resource_manager.downloads_folder

        if self._course_viewer_container.selected_course is not None:
            directory /= self._qcanvas.resource_manager.course_folder_name(
                self._course_viewer_container.selected_course
            )

        directory.mkdir(parents=True, exist_ok=True)

        QDesktopServices.openUrl(file_url(directory))

    @asyncSlot()
    async def _clear_new_items(self) -> None:
        await self._reload(empty_receipt())



from PySide6.QtCore import Slot, QUrl
from PySide6.QtWidgets import QWidget
from qasync import asyncSlot

from qcanvas.backend_connectors import FrontendResourceManager
from .qml_bridge_types import Attachment
from .qml_pane import QmlPane
from libqcanvas import db
import logging

_logger = logging.getLogger(__name__)


class AttachmentsPane(QmlPane):
    def __init__(
        self, downloader: FrontendResourceManager, parent: QWidget | None = None
    ):
        super().__init__(qml_path=QUrl("qrc:/qml/AttachmentsPane.qml"), parent=parent)

        self._original_dock_name = None
        self._downloader = downloader
        self._files: dict[str, db.Resource] = {}
        self._qfiles: dict[str, Attachment] = {}
        self.ctx["submission_files"] = []
        self.load_view()

        self._downloader.download_finished.connect(self._download_updated)
        self._downloader.download_failed.connect(self._download_updated)

    def clear_files(self):
        self.ctx["submission_files"] = []
        self._files.clear()
        self._qfiles.clear()

    def load_files(self, files: list[db.Resource]):
        qfiles = []

        if self._original_dock_name is None:
            self._original_dock_name = self.parent().windowTitle()

        self.parent().setWindowTitle(f"{self._original_dock_name} ({len(files)})")

        for file in files:
            qfile = Attachment(
                file_name=file.file_name,
                resource_id=file.id,
                download_state=file.download_state,
            )
            qfile.opened.connect(self._on_attachment_opened)
            qfiles.append(qfile)

            self._qfiles[file.id] = qfile
            self._files[file.id] = file

        self.ctx["submission_files"] = qfiles

    @asyncSlot(str)
    async def _on_attachment_opened(self, resource_id: str) -> None:
        if resource_id in self._files:
            await self._downloader.download_and_open(self._files[resource_id])
        else:
            _logger.warning(
                "User opened an attachment that doesn't belong to any comment! id=%s",
                resource_id,
            )

    @Slot(db.Resource)
    def _download_updated(self, resource: db.Resource) -> None:
        if resource.id in self._files:
            self._qfiles[resource.id].download_state = resource.download_state



import logging

from PySide6.QtCore import Slot
from PySide6.QtGui import QAction, QActionGroup
from PySide6.QtWidgets import QMenu

from qcanvas import icons
from qcanvas.theme import app_theme
import qcanvas.settings as settings

_logger = logging.getLogger(__name__)


class _ThemeAction(QAction):
    def __init__(self, theme_text: str, theme_name: str, parent: QMenu):
        super().__init__(theme_text, parent)
        self._theme_name = theme_name
        self.setCheckable(True)
        self.setChecked(self._theme_name == settings.ui.theme)
        self.triggered.connect(self._change_theme)

    @Slot()
    def _change_theme(self) -> None:
        settings.ui.theme = self._theme_name
        app_theme.theme = self._theme_name


class ThemeSelectionMenu(QMenu):
    def __init__(self, parent: QMenu):
        super().__init__("Theme", parent)

        action_group = QActionGroup(self)

        auto_theme = _ThemeAction("Auto", "auto", self)
        light_theme = _ThemeAction("Light", "light", self)
        dark_theme = _ThemeAction("Dark", "dark", self)
        native_theme = _ThemeAction("Native (requires restart)", "native", self)

        select_theme_actions = [auto_theme, light_theme, dark_theme, native_theme]

        self.addActions(select_theme_actions)
        self.setIcon(icons.options.theme)

        for selection_action in select_theme_actions:
            action_group.addAction(selection_action)



import logging
from asyncio import Lock
from dataclasses import dataclass

from libqcanvas.task_master import TaskID
from PySide6.QtWidgets import QProgressBar, QStatusBar
from qasync import asyncSlot

from qcanvas.backend_connectors import task_master

_logger = logging.getLogger(__name__)


@dataclass
class _TaskProgress:
    current: int
    total: int


class StatusBarProgressDisplay(QStatusBar):
    def __init__(self):
        super().__init__()
        self._lock = Lock()
        self._tasks: dict[TaskID, _TaskProgress] = {}

        self._progress_bar = self._setup_progress_bar()
        self.addPermanentWidget(self._progress_bar)

        task_master.task_progress.connect(self._on_task_progress)
        task_master.task_failed.connect(self._on_task_failed)

        self.showMessage("Ready", 5000)

    def _setup_progress_bar(self) -> QProgressBar:
        bar = QProgressBar()
        bar.setTextVisible(True)
        bar.hide()
        return bar

    @asyncSlot(TaskID, int, int)
    async def _on_task_progress(
        self, task_id: TaskID, current: int, total: int
    ) -> None:
        _logger.debug("Progress %s: %i/%i", task_id, current, total)

        async with self._lock:
            if task_id not in self._tasks:
                self._add_task(task_id, current, total)
            if current == total and total != 0:
                self._remove_task(task_id)
            else:
                self._update_task(task_id, current, total)

        await self._update_task_status()

    def _update_task(self, task_id: TaskID, current: int, total: int) -> None:
        _logger.debug("Update %s", task_id)
        task = self._tasks[task_id]
        task.current = current
        task.total = total

    @asyncSlot(TaskID, object)
    async def _on_task_failed(self, task_id: TaskID, context: object) -> None:
        _logger.info("%s failed", task_id)

        async with self._lock:
            self._remove_task(task_id)

            if self._has_no_tasks:
                self._progress_bar.hide()

        self.showMessage(f"Failed: {task_id.step_name}", 5000)

    async def _update_task_status(self) -> None:
        _logger.debug("Tasks: %s", self._tasks)
        async with self._lock:
            if self._has_no_tasks:
                self._show_done()
            else:
                tasks = list(self._tasks.items())

                if self._has_single_task:
                    self._show_single_task_progress(tasks[0])
                else:
                    self._show_multiple_tasks_progress(tasks)

    def _show_done(self) -> None:
        _logger.info("Finished tasks. Tasks: %s", self._tasks)
        self.showMessage("Done", 5000)
        self._progress_bar.hide()

    def _show_single_task_progress(self, task: tuple[TaskID, _TaskProgress]) -> None:
        _logger.debug("Single task %s", task)
        id, progress = task

        self._show_progress(progress)
        self.showMessage(id.step_name)

    def _show_multiple_tasks_progress(
        self, tasks: list[tuple[TaskID, _TaskProgress]]
    ) -> None:
        _logger.debug("Multiple tasks %s", tasks)
        self.showMessage(
            f"{len(tasks)} tasks in progress - {', '.join([task[0].step_name for task in tasks])}"
        )
        self._show_progress(self._calculate_progress(tasks))

    def _calculate_progress(
        self, tasks: list[tuple[TaskID, _TaskProgress]]
    ) -> _TaskProgress:
        # Task progresses are floats from 0 to 1, multiplier is used to turn them into ints
        multiplier = 1000
        current_sum = 0
        total_sum = 0

        for _, task in tasks:
            if task.total != 0:
                current_sum += (task.current / task.total) * multiplier

            total_sum += multiplier

        _logger.debug(
            "%s tasks, current=%i, total=%i", len(tasks), int(current_sum), total_sum
        )

        return _TaskProgress(int(current_sum), total_sum)

    def _show_progress(self, progress: _TaskProgress) -> None:
        self._progress_bar.setMaximum(progress.total)
        self._progress_bar.setValue(progress.current)

        if progress.total != 0:
            self._progress_bar.setFormat(
                f"{(progress.current / progress.total) * 100:.0f}%"
            )
        else:
            self._progress_bar.setFormat("")

        self._progress_bar.show()

    def _add_task(self, task: TaskID, current: int, total: int) -> None:
        self._tasks[task] = _TaskProgress(current, total)
        _logger.info("Added task %s", task)
        _logger.debug("Tasks: %s", self._tasks)

    def _remove_task(self, task: TaskID) -> None:
        self._tasks.pop(task, None)
        _logger.info("Removed task %s", task)
        _logger.debug("Tasks: %s", self._tasks)

    @property
    def _has_single_task(self) -> bool:
        return len(self._tasks) == 1

    @property
    def _has_many_tasks(self) -> bool:
        return len(self._tasks) > 1

    @property
    def _has_no_tasks(self) -> bool:
        return len(self._tasks) == 0



import logging
from typing import Sequence

from libqcanvas import db
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QHeaderView

from qcanvas import icons
from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.tree_widget_data_item import TreeWidgetDataItem
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class AssignmentTree(ContentTree[db.Course]):
    def __init__(self, course_id: str):
        super().__init__(
            tree_name=f"course.{course_id}.assignment_groups",
            emit_selection_signal_for_type=db.Assignment,
        )

        self.ui_setup(
            header_text=["Assignments", "Weight"],
            indentation=15,
            max_width=350,
            min_width=150,
        )

        self.set_columns_resize_mode(
            [QHeaderView.ResizeMode.Stretch, QHeaderView.ResizeMode.ResizeToContents]
        )

    def create_tree_items(
        self, course: db.Course, sync_receipt: SyncReceipt
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for assignment_group in course.assignment_groups:  # type: db.AssignmentGroup
            if len(assignment_group.assignments) == 0:
                continue

            assignment_group_widget = self._create_assignment_group_widget(
                assignment_group
            )

            widgets.append(assignment_group_widget)
            for assignment in assignment_group.assignments:  # type: db.Assignment
                assignment_widget = self._create_assignment_widget(
                    assignment, sync_receipt
                )

                assignment_group_widget.addChild(assignment_widget)

        return widgets

    def _create_assignment_group_widget(
        self, assignment_group: db.AssignmentGroup
    ) -> MemoryTreeWidgetItem:
        assignment_group_widget = MemoryTreeWidgetItem(
            id=assignment_group.id,
            data=assignment_group,
            strings=[assignment_group.name, f"{assignment_group.group_weight}%"],
        )

        assignment_group_widget.setFlags(Qt.ItemFlag.ItemIsEnabled)
        assignment_group_widget.setIcon(0, icons.tree_items.module)

        return assignment_group_widget

    def _create_assignment_widget(
        self, assignment: db.Assignment, sync_receipt: SyncReceipt
    ) -> TreeWidgetDataItem:
        assignment_widget = TreeWidgetDataItem(
            id=assignment.id, data=assignment, strings=[assignment.name]
        )

        assignment_widget.setFlags(
            Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        )
        assignment_widget.setIcon(0, icons.tree_items.assignment)

        if sync_receipt.was_updated(assignment):
            self.mark_as_unseen(assignment_widget)

        return assignment_widget



import logging
from pathlib import Path

from aiofile import async_open
from pydantic import BaseModel, RootModel, Field, ValidationError

from qcanvas.util import paths

_logger = logging.getLogger(__name__)


class CourseConfigData(BaseModel):
    nickname: str | None = Field(default=None)

    async def save(self) -> None:
        await course_configs.save()


_CourseConfigurations = RootModel[dict[str, CourseConfigData]]


class _CourseConfig:
    def __init__(self):
        self._root_model = self._load_root_model()

    def _load_root_model(self) -> _CourseConfigurations:
        if self._storage_path.exists():
            try:
                return _CourseConfigurations.model_validate_json(
                    self._storage_path.read_text()
                )
            except ValidationError as e:
                _logger.error("Failed to load course configs", exc_info=e)

        return _CourseConfigurations({})

    async def save(self) -> None:
        async with async_open(self._storage_path, "wt") as file:
            await file.write(self._root_model.model_dump_json(indent=4))

    @property
    def _storage_path(self) -> Path:
        return paths.config_storage() / "course_settings.json"

    def __getitem__(self, item: str) -> CourseConfigData:
        if item in self._root_model.root:
            return self._root_model.root[item]
        else:
            new_config = CourseConfigData()
            self._root_model.root[item] = new_config
            return new_config


course_configs = _CourseConfig()



import logging
from dataclasses import dataclass

from libqcanvas import db
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtCore import Slot
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QTabWidget, QVBoxLayout, QWidget

from qcanvas import icons
from qcanvas.ui.course_viewer.tabs.assignment_tab import AssignmentTab
from qcanvas.ui.course_viewer.tabs.mail_tab import MailTab
from qcanvas.ui.course_viewer.tabs.page_tab import PageTab
from qcanvas.util.layouts import layout
import qcanvas.util.ui_tools as ui

_logger = logging.getLogger(__name__)


@dataclass
class _Tab:
    icon: QIcon
    highlighted_icon: QIcon
    update_type: type


class CourseViewer(QWidget):
    def __init__(
        self,
        course: db.Course,
        downloader: ResourceManager,
        *,
        sync_receipt: SyncReceipt,
    ):
        super().__init__()
        # todo this is a mess. there are several other messes like this too, do they all have to be a mess?
        self._course_id = course.id
        self._previous_tab_index = 0

        self._course_label = ui.label(
            course.name, font=ui.font(point_size=13, bold=True), allow_truncation=True
        )

        self._pages_tab = PageTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=sync_receipt,
        )
        self._assignments_tab = AssignmentTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=sync_receipt,
        )
        self._mail_tab = MailTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=sync_receipt,
        )
        # self._files_tab = FileTab.create_from_receipt(
        #     course=course,
        #     downloader=downloader,
        #     sync_receipt=sync_receipt,
        # )

        self._tab_widget = QTabWidget()
        self._tabs: dict[int, _Tab] = {}

        # self._setup_tab(
        #     name="Files",
        #     widget=self._files_tab,
        #     icon=icons.tabs.pages,
        #     highlighted_icon=icons.tabs.pages_new_content,
        # )
        self._PAGES_TAB = self._set_up_tab(
            name="Pages",
            widget=self._pages_tab,
            icon=icons.tabs.pages,
            highlighted_icon=icons.tabs.pages_new_content,
            content_update_key=db.Page,
        )
        self._ASSIGNMENTS_TAB = self._set_up_tab(
            name="Assignments",
            widget=self._assignments_tab,
            icon=icons.tabs.assignments,
            highlighted_icon=icons.tabs.assignments_new_content,
            content_update_key=db.Assignment,
        )
        self._MAIL_TAB = self._set_up_tab(
            name="Mail",
            widget=self._mail_tab,
            icon=icons.tabs.mail,
            highlighted_icon=icons.tabs.mail_new_content,
            content_update_key=db.Message,
        )
        # self._tabs.addTab(QLabel("Not implemented"), "Panopto")  # The meme lives on!

        self.setLayout(layout(QVBoxLayout, self._course_label, self._tab_widget))
        self._tab_widget.currentChanged.connect(self._tab_changed)
        self._highlight_tabs(sync_receipt)

    def _set_up_tab(
        self,
        *,
        widget: QWidget,
        icon: QIcon,
        highlighted_icon: QIcon,
        name: str,
        content_update_key: type,
    ) -> int:
        index = self._tab_widget.addTab(widget, icon, name)
        self._tabs[index] = _Tab(icon, highlighted_icon, update_type=content_update_key)
        return index

    def reload(self, course: db.Course, *, sync_receipt: SyncReceipt) -> None:
        # self._files_tab.reload(course, sync_receipt=sync_receipt)
        self._pages_tab.reload(course, sync_receipt=sync_receipt)
        self._assignments_tab.reload(course, sync_receipt=sync_receipt)
        self._mail_tab.reload(course, sync_receipt=sync_receipt)
        self._highlight_tabs(sync_receipt)

    @Slot(int)
    def _tab_changed(self, index: int) -> None:
        _logger.debug(f"Index = {index}")
        if index != -1:
            _logger.debug(f"Previous tab = {self._previous_tab_index}")
            self._set_tab_highlight(self._previous_tab_index, False)
            self._previous_tab_index = index

    def _highlight_tabs(self, sync_receipt: SyncReceipt) -> None:
        updates = sync_receipt.updates_by_course.get(self._course_id, None)

        for tab_index, tab in enumerate(self._tabs.values()):
            self._set_tab_highlight(
                tab_index, updates is not None and updates[tab.update_type] is not None
            )

    def _set_tab_highlight(self, tab_index: int, highlighted: bool) -> None:
        tab = self._tabs[tab_index]
        self._tab_widget.setTabIcon(
            tab_index, tab.highlighted_icon if highlighted else tab.icon
        )



import logging
from typing import override

from PySide6.QtCore import Qt
from libqcanvas import db
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from libqcanvas.util import as_local
from PySide6.QtWidgets import QLabel, QLayout, QMainWindow, QDockWidget

from qcanvas.backend_connectors import FrontendResourceManager
from qcanvas.ui.course_viewer.tabs.content_tab import ContentTab
from qcanvas.ui.course_viewer.tabs.mail_tab.mail_tree import MailTree
from qcanvas.ui.course_viewer.tabs.constants import date_strftime_format
import qcanvas.util.ui_tools as ui
from qcanvas.ui.qml_components import AttachmentsPane

_logger = logging.getLogger(__name__)


# todo maybe update has_been_read? probably not the responsibility of this class though
class MailTab(ContentTab):
    def __init__(
        self,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: FrontendResourceManager,
    ):
        self._main_container = QMainWindow()

        super().__init__(
            explorer=MailTree.create_from_receipt(course, sync_receipt=sync_receipt),
            title_placeholder_text="No mail selected",
            downloader=downloader,
        )

        self._main_container.setCentralWidget(self._viewer)
        self._files_pane = AttachmentsPane(downloader)
        self._files_dock = ui.dock_widget(
            widget=self._files_pane,
            title="Attachments",
            name="attachments",
            min_size=ui.size(150, 100),
            features=QDockWidget.DockWidgetFeature.DockWidgetMovable,
        )
        self._main_container.addDockWidget(
            Qt.DockWidgetArea.TopDockWidgetArea, self._files_dock
        )
        self._date_sent_label = QLabel("")
        self._sender_label = QLabel("")

        self.enable_info_grid()

    @override
    def _setup_layout(self) -> None:
        super()._setup_layout()
        self.content_grid.replaceWidget(
            self._viewer,
            self._main_container,
        )

    def setup_info_grid(self) -> QLayout:
        return ui.form_layout(
            {"From": self._sender_label, "Date": self._date_sent_label},
        )

    def update_info_grid(self, mail: db.Message) -> None:
        self._date_sent_label.setText(
            as_local(mail.creation_date).strftime(date_strftime_format)
        )
        self._sender_label.setText(mail.sender_name)

        if mail.attachments:
            self._files_pane.load_files(mail.attachments)
            self._files_dock.show()
        else:
            self._files_dock.hide()

    @override
    def _show_blank(self) -> None:
        super()._show_blank()
        self._files_dock.hide()



import logging
import os
from pathlib import Path

import cachetools
import platformdirs
from PySide6.QtCore import QSettings

from qcanvas.util.runtime import (
    is_running_as_compiled,
    is_running_as_flatpak,
    is_running_portable,
)

_logger = logging.getLogger(__name__)


def data_storage() -> Path:
    return root()


def config_storage() -> Path:
    if is_running_portable:
        return root()
    else:
        path = platformdirs.user_config_path("QCanvasTeam", "QCanvas")
        path.mkdir(parents=True, exist_ok=True)
        return path


def client_settings() -> QSettings:
    if is_running_portable:
        return QSettings("QCanvas.ini", QSettings.Format.IniFormat)
    else:
        return QSettings("QCanvasTeam", "QCanvas")


@cachetools.cached(cachetools.LRUCache(maxsize=1))
def root() -> Path:
    root_path = Path()

    if is_running_as_flatpak:
        # Flatpak does not support portable mode
        root_path = Path(os.environ["XDG_DATA_HOME"])
    elif not is_running_portable and is_running_as_compiled:
        root_path = platformdirs.user_data_path("QCanvasReborn", "QCanvasTeam")
    elif not is_running_portable:
        _logger.warning("Don't know how we're being run? Are you running from source?")

    print("Root path", root_path.absolute())
    _logger.debug("Root path %s", root_path.absolute())
    return root_path



from .qml import rc_qml  # Do not remove!! Loads the qml resources
from .comments_pane import CommentsPane
from .attachments_pane import AttachmentsPane
from .qml_bridge_types import Attachment, Comment



import logging
from typing import Sequence

from libqcanvas import db
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtCore import Qt

from qcanvas import icons
from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.tree_widget_data_item import TreeWidgetDataItem
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class PageTree(ContentTree[db.Course]):
    def __init__(self, course_id: str):
        super().__init__(
            tree_name=f"course.{course_id}.modules",
            emit_selection_signal_for_type=db.Page,
        )

        self.ui_setup(
            header_text="Content", indentation=15, max_width=300, min_width=150
        )

    def create_tree_items(
        self, course: db.Course, sync_receipt: SyncReceipt
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for module in course.modules:  # type: db.Module
            if len(module.pages) == 0:
                continue

            module_widget = self._create_module_widget(module)
            widgets.append(module_widget)

            for page in module.pages:  # type: db.ModulePage
                page_widget = self._create_page_widget(page, sync_receipt)
                module_widget.addChild(page_widget)

        return widgets

    def _create_module_widget(self, module: db.Module) -> MemoryTreeWidgetItem:
        module_widget = MemoryTreeWidgetItem(
            id=module.id, data=module, strings=[module.name]
        )
        module_widget.setFlags(Qt.ItemFlag.ItemIsEnabled)
        module_widget.setIcon(0, icons.tree_items.module)

        return module_widget

    def _create_page_widget(
        self, page: db.Page, sync_receipt: SyncReceipt
    ) -> TreeWidgetDataItem:
        page_widget = TreeWidgetDataItem(id=page.id, data=page, strings=[page.name])
        page_widget.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
        page_widget.setIcon(0, icons.tree_items.page)

        if sync_receipt.was_updated(page):
            self.mark_as_unseen(page_widget)

        return page_widget



import logging
from pathlib import Path

from PySide6.QtCore import QUrl

_logger = logging.getLogger(__name__)


def file_url(path: Path) -> QUrl:
    return QUrl.fromLocalFile(str(path.absolute()))



from typing import Any, Sequence

from PySide6.QtCore import QSize
from PySide6.QtGui import QAction, QIcon, QKeySequence, QPixmap, QFont
from PySide6.QtWidgets import (
    QMenu,
    QSizePolicy,
    QLabel,
    QWidget,
    QFormLayout,
    QDockWidget,
    QLayout,
    QHBoxLayout,
    QVBoxLayout,
)


def font(*, point_size: float | int | None = None, bold: bool | None = None) -> QFont:
    _font = QFont()

    if point_size is not None:
        if isinstance(point_size, int):
            _font.setPointSize(point_size)
        elif isinstance(point_size, float):
            _font.setPointSizeF(point_size)
        else:
            raise TypeError("point_size")

    if bold is not None:
        _font.setBold(bold)

    return _font


_bold = font(bold=True)


def label(text: str, *, font: QFont = None, allow_truncation: bool = False) -> QLabel:
    _label = QLabel(text)

    if font is not None:
        _label.setFont(font)

    if allow_truncation:
        _label.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Fixed)

    return _label


def bold_label(text: str, *, allow_truncation: bool = False) -> QLabel:
    return label(text, allow_truncation=allow_truncation, font=_bold)


def form_layout(
    rows: dict[str, QWidget], label_font: QFont | None = _bold, label_suffix: str = ":"
) -> QFormLayout:
    layout = QFormLayout()

    for name, widget in rows.items():
        if label_font:
            layout.addRow(label(name + label_suffix, font=label_font), widget)
        else:
            layout.addRow(name + label_suffix, widget)

    return layout


def size(width: int, height: int):
    return QSize(width, height)


def dock_widget(
    *,
    widget: QWidget,
    title: str | None = None,
    hide: bool = False,
    name: str | None = None,
    min_size: QSize | None = None,
    features: QDockWidget.DockWidgetFeature | None = None,
    parent: QWidget | None = None,
) -> QDockWidget:
    dock = QDockWidget(title, parent)
    dock.setWidget(widget)

    if name is not None:
        dock.setObjectName(name)

    if min_size is not None:
        dock.setMinimumSize(min_size)

    if features is not None:
        dock.setFeatures(features)

    if hide:
        dock.hide()

    return dock


def widget(layout: QLayout) -> QWidget:
    widget = QWidget()
    widget.setLayout(layout)
    return widget


def hbox(*items: QWidget | QLayout) -> QHBoxLayout:
    _add_layout_items(layout := QHBoxLayout(), items)
    return layout


def vbox(*items: QWidget | QLayout) -> QVBoxLayout:
    _add_layout_items(layout := QVBoxLayout(), items)
    return layout


def _add_layout_items(layout: QLayout, items: Sequence[QWidget | QLayout]):
    for item in items:
        if isinstance(item, QLayout):
            layout.addItem(item)
        else:
            layout.addWidget(item)


def create_qaction(
    *,
    name: str,
    shortcut: QKeySequence | None = None,
    parent: QMenu = None,
    triggered: Any = None,
    checkable: bool | None = None,
    checked: bool | None = None,
    icon: QIcon | QPixmap | None = None,
) -> QAction:
    action = QAction(name)

    if shortcut is not None:
        action.setShortcut(shortcut)

    if parent is not None:
        action.setParent(parent)
        parent.addAction(action)

    if triggered is not None:
        action.triggered.connect(triggered)

    if checkable is not None:
        action.setCheckable(checkable)

        if checked is not None:
            action.setChecked(checked)

    if icon is not None:
        action.setIcon(icon)

    return action



import logging
from abc import ABCMeta
from pathlib import Path

from PySide6.QtGui import QDesktopServices
from libqcanvas import db
from libqcanvas.database import QCanvasDatabase
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.resources.extracting.extractors import Extractors
from PySide6.QtCore import QObject, Signal

from qcanvas.util.qurl_util import file_url

_logger = logging.getLogger(__name__)


class _Meta(type(QObject), ABCMeta): ...


class FrontendResourceManager(QObject, ResourceManager, metaclass=_Meta):
    download_finished = Signal(db.Resource)
    download_failed = Signal(db.Resource)
    download_progress = Signal(db.Resource, int, int)

    def __init__(
        self,
        database: QCanvasDatabase,
        download_dest: Path,
        extractors: Extractors = Extractors(),
    ):
        super().__init__(
            database=database, download_dest=download_dest, extractors=extractors
        )

    async def download_and_open(self, resource: db.Resource) -> None:
        await self.download(resource)
        resource_path = file_url(self.resource_download_location(resource))
        QDesktopServices.openUrl(resource_path)

    def on_download_progress(
        self, resource: db.Resource, current: int, total: int
    ) -> None:
        self.download_progress.emit(resource, current, total)

    def on_download_failed(self, resource: db.Resource) -> None:
        self.download_failed.emit(resource)

    def on_download_finished(self, resource: db.Resource) -> None:
        self.download_finished.emit(resource)



import asyncio
import logging
from pathlib import Path

from aiofile import async_open

from qcanvas.util import paths

_logger = logging.getLogger(__name__)


class TreeMemory:
    def __init__(self, tree_name: str):
        self._tree_name = tree_name
        self._loaded = False
        self._collapsed_items: set[str] = set()

    def load(self, force: bool = False):
        if force or not self._loaded:
            self._loaded = True

            if not self._storage_path.exists():
                # Nothing to do
                return

            # fixme this blocks the event loop, but using aiofile will require significant changes to other
            #  parts of the code to accommodate these methods now being async.
            #  this will really only have any noticeable effect on slow disks, and only ever happens once anyway.
            with open(self._storage_path, "rt") as file:
                lines = file.read().splitlines()
                _logger.debug("Tree % loaded %s", self._tree_name, lines)
                self._collapsed_items.update(lines)

    async def save(self):
        assert self._loaded, "Memory is not loaded yet"

        async with async_open(self._storage_path, "wt") as file:
            await file.write("\n".join(self._collapsed_items))

    def expanded(self, node_id: str) -> None:
        self.set_expanded(node_id, True)

    def collapsed(self, node_id: str) -> None:
        self.set_expanded(node_id, False)

    def set_expanded(self, node_id: str, expanded: bool) -> None:
        self.load()

        if expanded and node_id in self._collapsed_items:
            self._collapsed_items.remove(node_id)
        else:
            self._collapsed_items.add(node_id)

        # hack: avoid blocking the event loop
        asyncio.create_task(self.save())

    @property
    def collapsed_ids(self) -> set[str]:
        assert self._loaded, "Memory not loaded yet"
        return self._collapsed_items

    @property
    def _storage_path(self) -> Path:
        path = (
            paths.data_storage()
            / "tree_state"
            / f"{self._tree_name}_collapsed_items.txt"
        )
        path.parent.mkdir(parents=True, exist_ok=True)
        return path



import logging

import qcanvas.settings as settings
from qcanvas.util.url_checker import is_url

_logger = logging.getLogger(__name__)


def needs_setup() -> bool:
    if not settings.client.panopto_disabled and not is_url(settings.client.panopto_url):
        return True
    elif not is_url(settings.client.canvas_url):
        return True
    elif len(settings.client.canvas_api_key) == 0:
        return True
    else:
        return False



from PySide6.QtCore import Property, QObject, Signal
from PySide6.QtQml import ListProperty, QmlElement
from libqcanvas.database.tables import ResourceDownloadState


QML_IMPORT_NAME = "QCanvas"
QML_IMPORT_MAJOR_VERSION = 1


@QmlElement
class Attachment(QObject):
    file_name_changed = Signal()
    resource_id_changed = Signal()
    download_state_changed = Signal()

    # Emitted by AttachmentsListDelegate when the user clicks on an attachment
    opened = Signal(str)

    def __init__(
        self,
        file_name: str,
        resource_id: str,
        download_state: ResourceDownloadState,
        parent: QObject | None = None,
    ):
        super().__init__(parent)
        self._file_name = file_name
        self._resource_id = resource_id
        self._download_state = download_state.name

    @Property(str, notify=file_name_changed)
    def file_name(self) -> str:
        return self._file_name

    @Property(str, notify=resource_id_changed)
    def resource_id(self) -> str:
        return self._resource_id

    @Property(str, notify=download_state_changed)
    def download_state(self) -> str:
        return self._download_state

    @download_state.setter
    def download_state(self, value: ResourceDownloadState):
        if value.name != self._download_state:
            self._download_state = value.name
            self.download_state_changed.emit()


@QmlElement
class Comment(QObject):
    body_changed = Signal()
    author_changed = Signal()
    date_changed = Signal()
    attachments_changed = Signal()

    def __init__(
        self,
        body: str,
        author: str,
        date: str,
        attachments: list[Attachment],
        parent: QObject | None = None,
    ):
        super().__init__(parent)

        self._body = body
        self._author = author
        self._date = date
        self._attachments = attachments

    @Property(str, notify=body_changed)
    def body(self) -> str:
        return self._body

    @Property(str, notify=date_changed)
    def date(self) -> str:
        return self._date

    @Property(str, notify=author_changed)
    def author(self) -> str:
        return self._author

    def attachment(self, n) -> Attachment:
        return self._attachments[n]

    def attachment_count(self) -> int:
        return len(self._attachments)

    # You must set `count`, `at` and `notify` EXPLICTLY (even if you name them according to the examples, which examples are all inconsistent and wrong).
    # Qt?? Are you ok???
    # Oh and you have to use `.length` instead of `.count` on the QML side because javascript.
    attachments = ListProperty(
        Attachment, count=attachment_count, at=attachment, notify=attachments_changed
    )



import asyncio
import logging

from PySide6.QtCore import Qt
from libqcanvas import db
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from PySide6.QtWidgets import QMessageBox, QWidget

import qcanvas.settings as settings

_logger = logging.getLogger(__name__)


async def download_new_resources(
    *,
    all_resources: dict[str, db.Resource],
    receipt: SyncReceipt,
    downloader: ResourceManager,
    parent_window: QWidget,
) -> None:
    resources_to_download = []

    for file_id in receipt.updates[db.Resource]:
        resource = all_resources[file_id]

        if _should_auto_download_resource(resource, resource_manager=downloader):
            resources_to_download.append(resource)

    resource_count = len(resources_to_download)

    if resource_count == 0:
        return
    elif resource_count > 20:
        msg = _confirmation_messagebox(resource_count, parent_window)
        msg.show()
        msg.accepted.connect(
            lambda: asyncio.get_running_loop().create_task(
                downloader.batch_download(resources_to_download),
            ),
            Qt.ConnectionType.SingleShotConnection,
        )
    else:
        await downloader.batch_download(resources_to_download)


def _should_auto_download_resource(
    resource: db.Resource, resource_manager: ResourceManager
) -> bool:
    return settings.client.download_new_videos or not resource_manager.is_video(
        resource
    )


def _confirmation_messagebox(resource_count: int, parent: QWidget) -> QMessageBox:
    return QMessageBox(
        QMessageBox.Icon.Question,
        "Many files to download",
        f"You are about to download {resource_count} new files.\n"
        "Do you want to continue?",
        QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        parent=parent,
    )
